
(* this file is based on lksdt, modified *)

structure ipldt = struct val thy = theory "ipldt" end ;
use_legacy_bindings ipldt.thy ;

(*** proofs for ipl ***)

(** inductive proof of weakening **)

val () = qed_goal_spec_mp "ljss_wk_adm" ipldt.thy "wk_adm_antec ljss"
  (fn _ => [ (simp_tac (esimps [ljss_iextrs, wk_adm_antec_iextrs]) 1) ]) ;

val _ = bind_thm ("ljw", ljss_wk_adm RS wk_adm_antecD_alt) ;

val ljwl = rewrite_rule [drr_empty] ljw ;

val lj_ax = simplify (esimps [extend_def]) (ljne.axiom RS ljss_extI2) ;
val lj_axd = fold_rule [drr_empty] 
  (transfer ipldt.thy lj_ax RS drl.singleI) ;

val () = qed_goalw_spec_mp "ljw_ax" ipldt.thy [mk_eq ljss_iextrs]
  "A :# X --> (X |- {#A#}) : derrec ljss {}"
  (fn _ => [ (Safe_tac), (rtac iextrs_lid_ax 1), Auto_tac ]) ;

val lj_axs = in_single RS ljw_ax ;

val () = qed_goal_spec_mp "lj_extle" ipldt.thy 
  "(ps, X |- Y) : ljne --> X <= Z --> \
    \ map (extend (Z - X |- 0)) ps : dersrec ljss qs --> \
    \ (Z |- Y) : derrec ljss qs"
  (fn _ => [ (strip_tac 1), (dtac ljss.extI 1), 
    rewtac (mk_eq pscmap_def), (datac drs.derI 1 1),
    (asm_full_simp_tac (esimps [extend_def]) 1)]) ;

val _ = qed_goal "zleif" HOL_Gen.thy
  "((0 :: nat) < (if A = fp then 1 else 0)) = (A = fp)" 
  (fn _ => [ (auto_tac (cis ipsfIs, esimps [split_if_asm])) ]) ;

(* OK TO HERE *)

(* not true since 
  (1) some rules have principal formula repeated in the premises
  (2) the left rules have the singletons on the right included in the rule
  need to look at why the result mattered, where it was used
  
  the property subfml_cp_prop is used in gen_ctr_adm,
  to prove contraction admissibility in general

  need corresponding general result

val () = qed_goal "lj_subfml_cp_prop" ipldt.thy 
  "Ball ljne (subfml_cp_prop ipsubfml)"
  (fn _ => [ (clarsimp_tac (cesimps [ms_mem_def, subfml_cp_prop.simps]) 1), 
    (safe_tac (ces [ljne.elims, ljir.elims, ljil.elims])),
    (ALLGOALS (clarsimp_tac (cesimps [zleif]))),
    (auto_tac (cis ipsfIs, esimps [split_if_asm])) ]) ;

    *)

(** to derive a sequent of the form seq : derrec ljss {} **) 

(* from logic.ML, but changed *)
(*Get subgoal i*)
local open Logic in
fun get_goal st i = nth (strip_imp_prems st, i-1)
  handle Subscript => error "get_goal: Goal number out of range";

(*reverses parameters for substitution*)
fun goal_params st i =
  let val gi = get_goal st i
      (* rev omitted from following line *)
      val rfrees = (map Free (rename_wrt_term gi (strip_params gi)))
  in (gi, rfrees) end;

fun concl_of_goal st i =
  let val (gi, rfrees) = goal_params st i
      val B = strip_assums_concl gi
  in subst_bounds (rfrees, B) end;
end ;

fun isFree (Free _) = true
  | isFree _ = false ;

val lks_lis = map (fold_rule [Zero_multiset_def]) 
  (lksil.intros @ lksir.intros) ;
val lks_nes = lks_lis RL lksne.intros ;

(** now prove invertibility **)

(** rules give derivable sequents **)

fun las_tacs seqstr _ = [ Clarify_tac 1,
  (dres_inst_tac [("flr1", seqstr)] 
    ([extrs_derl_le, refl RSN (2, extrs.I)] MRS subsetD) 1),
  Clarsimp_tac 1, (etac derl_trans 1),
  (forw_inst_tac [("ps1", "[]")] (thin_rl RS drl.dtCons RS mem_same) 1),
  (rtac drl.asmsI 2), (rtac refl 2), MSSG Full_simp_tac 2, 
  (rtac (drl.in_derl RS subsetD) 1),
  (eresolve_tac [ex_antE, ex_sucE] 1), 
  (clarsimp_tac (cesimps [extend_def]) 1),
  (etac (extrs_mono RS subsetD) 1), 
  (res_inst_tac [("flr", "x"), ("A1", "A")] (idrls.I RS extrs.I) 1),
  (simp_tac spsimps 1) ] ;

val () = qed_goal_spec_mp "log_thm_lem_ant" ipldt.thy 
  "(p # ps, c) : derl (extrs rls) --> idrls <= rls --> A :# antec p --> \
    \ pscmap (extend (0 |- {#A#})) (ps, c) : derl (extrs rls)" 
  (las_tacs "0 |- {#A#}") ;

val () = qed_goal_spec_mp "log_thm_lem_suc" ipldt.thy 
  "(p # ps, c) : derl (extrs rls) --> idrls <= rls --> A :# succ p --> \
    \ pscmap (extend ({#A#} |- 0)) (ps, c) : derl (extrs rls)" 
  (las_tacs "{#A#} |- 0") ;

val de_lem = rev_subsetD RS in_extrs RS 
  (transfer ipldt.thy (mk_mono' drl.in_derl));
val lks_nes_de = lks_nes RL [de_lem] ;
val lks_nes_de_ltls = lks_nes_de RL [log_thm_lem_ant, log_thm_lem_suc] ;

fun proc_ltl th = 
  let val th1 = full_simplify (esimps [in_gen_count, mset_map_union]) th ;
    val th2s = [refl, refl RS disjI1, refl RS disjI2] RLN (3, [th1]) ;
  in th2s end ;

(* selector theorems *)
fun pes th = ([th] RL [log_thm_lem_ant]) = [] ;

val all_ps = map zero_var_indexes (List.concat (map proc_ltl lks_nes_de_ltls)) ;
val all_ps_nps = List.filter pes all_ps ;
val all_ps_1p as [or1p, imp1p, and1p] = List.filter (not o pes) all_ps ;

(* where two premises, handle the second one *)
fun proc_ltl2 th = 
  let val th1 = full_simplify (esimps [in_gen_count]) th ;
    val th2 = refl RSN (4, th1)
      (* special case for imp-left rule *)
      handle _ => refl RS disjI1 RSN (4, th1) ;
  in rdp th2 end ;

val all_ps2 = map proc_ltl2 [or1p RS log_thm_lem_ant,
  imp1p RS log_thm_lem_suc, and1p RS log_thm_lem_suc ] ;

val all_log_thms = all_ps2 @ all_ps_nps ;

fun mk_lks th = tacsthm [ TRYALL (rtac lksne_idrls), rtac subset_refl 1] th ;
val lks_alts = map mk_lks all_log_thms ;
val lks_alts' = map (fcomm_all add_commute) lks_alts ;

val rstacs' = [(safe_tac (ces [lksne.elims])), (Full_simp_tac 1), 
  (TRYALL (dresolve_tac lks_concl_ilrc THEN' Force_tac)),
  (eresolve_tac [lksil.elims, lksir.elims] 1),
  (ALLGOALS (asm_full_simp_tac (esimps [extend_def]))), Safe_tac ] ;

val () = qed_goal_spec_mp "lks_rsr" ipldt.thy 
  "(ps, 0 |- {#A#}) : lksne --> flr = ({#A#} |- 0) --> \
    \ map (extend flr) ps : dersrec ljss {}"
  (fn _ => [ EVERY rstacs', rewtac ljss_def, rewtac drr_empty, 
    (TRYALL (resolve_tac lks_alts)),
    (TRYALL (resolve_tac lks_alts')) ]) ;
 
val () = qed_goal_spec_mp "lks_rsl" ipldt.thy 
  "(ps, {#A#} |- 0) : lksne --> flr = (0 |- {#A#}) --> \
    \ map (extend flr) ps : dersrec ljss {}"
  (fn _ => [ EVERY rstacs', rewtac ljss_def, rewtac drr_empty, 
    (TRYALL (resolve_tac lks_alts)),
    (TRYALL (resolve_tac lks_alts')) ]) ;

val rs1 = map (simplify (esimps [drs.all])) [lks_rsl, lks_rsr] ;
val rs2 = rs1 RL [image_subset_iff RS iffD1 RS bspec] ;
val rs3 = map (rewrite_rule [mk_eq ljss_extrs]) rs2 ;

val () = qed_goalw "lks_id_decomp" ipldt.thy [id_decomp_def] 
  "id_decomp (extrs lksne)"
  (fn _ => [ Safe_tac, (etac extrs.elims 1),
    (rtac lksne.elims 1), atac 1, Force_tac 1, Safe_tac, 
    (ftac lks_concl_ilc 1), (ftac lks_concl_irc 2),
    (ALLGOALS Clarsimp_tac), (ALLGOALS (eresolve_tac rs3)), (TRYALL atac),
    (ALLGOALS (case_tac "flr")),
    (ALLGOALS (clarsimp_tac (cesimps [extend_def]))),
    (ALLGOALS (forw_inst_tac 
      [("f", "%ms. (count ms xa, count ms A)")] arg_cong)),
    (ALLGOALS (full_simp_tac (esimps [] addsplits [split_if_asm]))),
    Safe_tac, 
    (ALLGOALS (full_simp_tac (HOL_ss addsimprocs pmgsps))) ]) ;
    
(* proof of lks_inv using generalised version *)
val () = qed_goalw_spec_mp "lks_inv" ipldt.thy [mk_eq ljss_extrs]
  "(ps, c) : lksne --> ce : derrec ljss {} --> (ALL plr. \
    \ ce = extend plr c --> map (extend plr) ps : dersrec ljss {})"
  (fn _ => [ Safe_tac, (rtac gen_inv' 1), atac 5, atac 4, 
    (rtac extrs.I 4), (rtac pscmap_def 5), (rtac singletonI 4),
    (rtac lks_id_decomp 3), (rtac lks_unique_concl 2),
    (rtac lks_iscrls 1) ]) ;

(* steps of proving lks_inv, useful for other rule sets *)
val lks_ds_lem = [lks_id_decomp, asm_rl, asm_rl, lksne.axiom, idrls.I] MRS 
  transfer ipldt.thy id_decompm_step1e ;

val lks_uc = meta_std ([unique_concl_def, lks_unique_concl] MRS defD1) ;

val _ = qed_goal_spec_mp "lks_inv_stepm_lem" ipldt.thy
  "irl : lksne --> drl : lksne --> psc : extrs {drl} --> \
    \ inv_stepm (extrs lksne) (extrs {irl}) psc"
  (fn _ => [ Clarify_tac 1, (etac lksne.elims 1), 
    (etac (surjective_pairing RS inv_stepm_empty) 1),
    (safe_tac (ces [lksne.elims])),
    (etac extrs.elims 1), (rtac lks_ds_lem 1), 
    Fast_tac 3, (etac lksne.ilI 1), (etac lksil_scrls 1), 
    (etac extrs.elims 3), (rtac lks_ds_lem 3), 
    Fast_tac 5, (etac lksne.irI 3), (etac lksir_scrls 3), 
    (ALLGOALS (EVERY' [(etac (refl RS inv_stepm_scrls RS inv_step_mono)),
      (REPEAT1 o eresolve_tac [lksil_scrls, lksir_scrls]),
      (eresolve_tac [extrs_il_lksne, extrs_ir_lksne] o incr),
      Clarsimp_tac, (rtac lks_uc), 
      (REPEAT1 o eresolve_tac [lksne.ilI, lksne.irI]) ])) ]) ;
      
val _ = qed_goal_spec_mp "lks_inv_stepm" ipldt.thy
  "psc : extrs lksne --> inv_stepm (extrs lksne) (extrs lksne) psc"
  (fn _ => [ (Clarify_tac 1),
    (rtac inv_stepm_UN' 1), (rtac extrs_UN_single 1), Clarify_tac 1, 
    (dtac (extrs_UN_single RS equalityD2') 1), Clarify_tac 1, 
    (rtac lks_inv_stepm_lem 1), Auto_tac ]) ;

val () = qed_goal "lks_inv_rl" ipldt.thy 
  "Ball (extrs lksne) (inv_rl (extrs lksne))"
  (fn _ => [ (Clarify_tac 1), (rtac gen_inv 1), atac 4,
    (rtac lks_id_decomp 3), (rtac lks_unique_concl 2),
    (rtac lks_iscrls 1) ]) ;

(* alternative proof *)
val () = qed_goal "lks_inv_rl" ipldt.thy 
  "Ball (extrs lksne) (inv_rl (extrs lksne))"
  (fn _ => [ (Clarify_tac 1), (rtac gen_inv_by_step 1), Clarify_tac 1,
    (rtac inv_step_m 1), atac 2, (etac lks_inv_stepm 1) ]);

(** admissibility of contraction **)

(* proof of ljss_ctr_adm using the generalised version *)
val () = qed_goal "lksne_ctr_adm" ipldt.thy 
  "ctr_adm (extrs lksne) A"
  (fn _ => [ (rtac gen_ctr_adm 1),
    (rtac wfp_ipsubfml 1), (rtac lks_iscrls 1),
    (rtac lks_subfml_cp_prop 1), (rtac lks_inv_rl 1) ]) ;

val () = bind_thm ("ljss_ctr_adm", fold_rule [ljss_def] lksne_ctr_adm) ;

val [ljss_ctr_alt, ljss_ctr_alt2, ljss_double_adm] = 
  [ljss_ctr_adm RS ballI] RL [ctr_alt, ctr_alt2, double_adm] ;
  
val ljss_double_eqv' = [ljss_ctr_adm, ljss_wk_adm] MRS double_eqv' ;
val ljss_double = ljss_double_eqv' RS iffD1 ;

val ljss_double_eqv = prove_goal ipldt.thy 
  "((X + X |- Y + Y) : derrec ljss {}) = ((X |- Y) : derrec ljss {})"
  (fn _ => [ (rtac trans 1), (rtac ljss_double_eqv' 2), Simp_tac 1]) ;

val _ = bind_thm ("wc_ljss", rotate_prems 1
  ([ljss_wk_adm, ljss_ctr_adm] MRS (allI RSN (2, wk_ctr_adm_singles)))) ;

val _ = bind_thm ("ljss_singles",
  seq_singles_idemp RS sym RS order_eq_refl RSN (2, wc_ljss)) ;

val _ = bind_thm ("ljss_singles_iff", tacsthm [ (etac ljss_singles 2),
  (etac lksw 1), (rtac seq_singles_le 1)] iffI) ;
  
val _ = qed_goal_spec_mp "lks_0_lem" ipldt.thy "r : lksil Un lksir --> \
    \ (EX ps. c : set ps & (ps, c) : extrs (multiples {r}))"
  (fn _ => [ (safe_tac (ces [lksil.elims, lksir.elims])),
    (ALLGOALS (rtac mult_0_lem)), Auto_tac ]) ;

(* former proof of this result, adapted from Jesse's proof 
val _ = qed_goal_spec_mp "ctr_adm_step_lksne" ipldt.thy
  "epsc : extrs lksne --> extrs lksne <= drls --> \
    \ Ball (extrs lksne) (inv_rl drls) --> \
    \ ctr_adm_step ipsubfml (derrec drls {}) epsc A"
  (fn _ => [ (safe_tac (ces [extrs.elims, lksne.elims])),
    (* axiom *)
    (rtac id_ctr_adm_step 1), (rtac idrls.intros 1), (Force_tac 1),
    (etac xt6 1), (rtac extrs_mono 1),
    (rtac (lksne.axiom RS singleton_subsetI) 1),
    (* lksil and lksir *)
    (REPEAT_DETERM_N 2 (EVERY [
      (rtac sc_ctr_adm_step 1), (rtac refl 1),
      (eresolve_tac [lksil_scrls, lksir_scrls] 1), (Force_tac 1),
      (rtac (lks_subfml_cp_prop RS bspec) 1),
      (eresolve_tac [lksne.ilI, lksne.irI] 1), (Clarsimp_tac 1),
      (etac bspec 1), 
      (dresolve_tac [extrs_il_lksne, extrs_ir_lksne] 1),
      (eatac subsetD 1 1),
      (dresolve_tac [extrs_il_lksne, extrs_ir_lksne] 1),
      (eatac subset_trans 1 1) ])) ]) ;
*)

val _ = bind_thm ("ctr_adm_step_lksne", 
  [asm_rl, lks_iscrls, asm_rl, lks_subfml_cp_prop] MRS gen_ctr_adm_step_inv) ;


