
structure lctr = struct val thy = theory "lctr" end ;
use_legacy_bindings lctr.thy ;

Delsimps [lext_concl_rl.simps] ;

val lctr_adm_def = rewrite_rule [ctr_adm_seq_def, lctr_admd_def] lctr_adm_def';
val lctr_admD = meta_std (lctr_adm_def RS defD1) ;
val lext_concl_def = rewrite_rule [mk_eq lext_concl_rl.simps] lext_concl_def';

val _ = qed_goal_spec_mp "iextrs_id" lctr.thy 
  "A :# X --> ([], X |- {#A#}) : iextrs idrls"
  (fn _ => [ (Clarify_tac 1), 
    (REPEAT (etac single_add_diff_exE 1)), Clarify_tac 1,
    (rtac (idrls.I RS iextrs.I) 1),  
    (simp_tac (esimps [extend_def]) 1), Fast_tac 1 ]) ;

(** admissibility of weakening **)
val _ = qed_goalw_spec_mp "wk_adm_antec_alt" lctr.thy 
  [wk_adm_antec_def', meta_eq_def]
  "wk_adm_antec rls == ALL X Y. (X |- Y) : derrec rls {} --> \
    \ (ALL X'. X <= X' --> (X' |- Y) : derrec rls {})"
  (fn _ => [ Safe_tac, 
    (eres_inst_tac [("x", "X |- Y")] allE 1), mp_tac 1,
    (dtac le_add_diff_ex 1), Clarsimp_tac 1,
    (case_tac "XY" 1), Clarify_tac 1,
    (etac allE2 1), (mp_tac 1), Simp_tac 1,
    (etac allE 1), (etac mp 1), (rtac le_plus1 1) ]) ;

val wk_adm_antec_alt' = meta_std (wk_adm_antec_alt RS defD1) ;

(*
(* to use a rule plus weakening in forward proof *)
val _ = qed_goalw_spec_mp "wk_adm_rule_sub" lctr.thy [wk_adm_alt]
  "wk_adm rls --> extrs {([p], c)} <= adm rls --> s : derrec rls {} --> \
    \ s - p + c : derrec rls {}"
  (fn _ => [ Safe_tac, (etac allE 1), mp_tac 1,
    (etac allE 1), (etac impE 1), (rtac le_mp 1),
    (dres_inst_tac [("flr1", "s - p")]
      (singletonI RS extrs_I RS rev_subsetD) 1),
    (clarsimp_tac (cesimps [extend_def']) 1), (etac thin_rl 1), 
    (etac admE 1), (etac impE 1), (simp_tac (esimps [drs.all]) 1), atac 1 ]) ;
*)

val _ = qed_goalw_spec_mp "wk_adm_lext_concl" lctr.thy [wk_adm_antec_def']
  "lext_concl rls --> wk_adm_antec rls"
  (fn _ => [ (rtac impI 1), (rtac allI 1), (rtac impI 1),
    (etac drs.inductr 1), Fast_tac 1,
    (rtac allI 1), (rewtac lext_concl_def),
    (datac bspec 1 1), Full_simp_tac 1, 
    (etac allE 1), (etac exE 1), (rtac drs.derI 1),
    (etac conjunct1 1), (rewtac (mk_eq drs.all)), Clarify_tac 1, 
    (datac allrel_ex 1 1), Clarify_tac 1, (datac bspec 1 1),
    (dres_inst_tac [("x", "X' - X")] spec 1),
    (dtac le_add_diff_eq 1), Asm_full_simp_tac 1 ]) ;

val lecmono = prove_goal lctr.thy 
  "A <= B --> lext_concl_rl A rl --> lext_concl_rl B rl" 
  (fn _ => [ (case_tac "rl" 1), Clarify_tac 1, 
    (full_simp_tac (esimps [lext_concl_rl.simps]) 1),
    (etac all_forward 1), (etac ex_forward 1), Fast_tac 1 ]) ;

val _ = bind_thm ("lext_concl_rl_mono", lecmono RS mp) ;
val _ = bind_thm ("lext_concl_rl_mono'", mk_mono' lext_concl_rl_mono) ;

val _ = qed_goalw_spec_mp "lext_concl_Un" lctr.thy [lext_concl_def']
  "lext_concl P --> lext_concl Q --> lext_concl (P Un Q)" 
  (fn _ => [ Safe_tac, (ALLGOALS (EVERY' [(datac bspec 1), 
    Full_simp_tac, (etac lext_concl_rl_mono'), Fast_tac])) ]) ;

val _ = qed_goalw "iextcs_lext_concl" lctr.thy [lext_concl_def]
  "lext_concl (iextcs rls)" 
  (fn _ => [ Safe_tac, (etac iextcs.elims 1), (Clarify_tac 1),
    (rtac exI 1), (rtac conjI 1),
    (res_inst_tac [("g1", "id")] (allrel_map' RS iffD1) 2),
    Clarsimp_tac 2, 
    (REPEAT_ALL_NEW (resolve_tac [exI, conjI, antec_succ_simp RS sym,
      pm_ge0_refl]) 2),
    (eres_inst_tac [("fl", "fl + A")] iextcs.intros 1),
    (simp_tac (esimps [id_def, map_id, extend_def, add_assoc]) 1) ]) ;

val _ = qed_goalw "iextrs_lext_concl" lctr.thy [lext_concl_def]
  "lext_concl (iextrs rls)" 
  (fn _ => [ Safe_tac, (etac iextrs.elims 1), (Clarify_tac 1),
    (rtac exI 1), (rtac conjI 1),
    (eres_inst_tac [("fl", "fl + A")] iextrs.intros 1),
    Clarsimp_tac 1, (rtac conjI 1), (rtac refl 1), 
    (simp_tac (esimps [extend_def, add_assoc]) 1),
    (clarsimp_tac (cesimps [allrel_map RS sym, extend_def]) 1),
    (case_tac "x" 1), Asm_simp_tac 1 ]) ;

AddIffs [iextrs_lext_concl, iextcs_lext_concl] ;

val _ = bind_thm ("iextrs_cs_lext_concl", 
  [iextrs_lext_concl, iextcs_lext_concl] MRS lext_concl_Un) ;

val _ = bind_thm ("wk_adm_iextrs_cs",
  iextrs_cs_lext_concl RS wk_adm_lext_concl) ;

val _ = qed_goalw "lidrls_adm_iff" lctr.thy [lid_adm_def]
  "lid_adm rls = (iextrs idrls <= adm rls)"
  (fn _ => [ (simp_tac (esimps adm.defs) 1), 
    (safe_tac (ces [iextrs.elims, idrls.elims])), 
    (clarsimp_tac (cesimps [extend_def, symmetric mins_def]) 1), 
    (dtac rel_subsetD 1), Clarsimp_tac 2, (etac mp 2),
    (rtac drs.dlNil 2), (rtac (idrls.I RS iextrs.I) 1), (Clarsimp_tac 1),  
    (rtac conjI 1), (rtac refl 2),
    (dtac (transfer lctr.thy single_le RS iffD2 RS 
      le_add_diff_eq) 1), atac 1 ]) ;

val _ = bind_thm ("lid_admD", meta_std (lid_adm_def RS defD1)) ;

val _ = qed_goalw_spec_mp "lid_adm_wk" lctr.thy 
  [lid_adm_def, wk_adm_antec_def]
  "idrls <= adm rls --> wk_adm_antec rls --> lid_adm rls"
  (fn _ => [ Safe_tac, (dtac (idrls.I RS rev_subsetD) 1),
    (dtac (drs.dlNil RSN (2, admD)) 1),
    (REPEAT_FIRST (eresolve_tac [allE, impE])), (atac 1), 
    (rtac succ.simps 1), 
    (dtac (transfer lctr.thy single_le RS iffD2 RS 
      le_add_diff_eq) 1),
    (etac (sds_plus_def RS def_imp_eq RS trans RSN (2, mem_same)) 1),
    Force_tac 1 ]) ;

val wk_adm_antec_iextrs = iextrs_lext_concl RS wk_adm_lext_concl ;

(* alternative proof *)
val _ = qed_goalw_spec_mp "wk_adm_antec_iextrs" lctr.thy [wk_adm_antec_def]
  "wk_adm_antec (iextrs rls)"
  (fn _ => [ (Safe_tac), (etac drs.inductr 1), Fast_tac 1,
    (case_tac "As" 1), Clarsimp_tac 1, 
    (dtac (iextrs_I RS (iextrs_iextrs RS equalityD1')) 1),
    (etac (mem_same RS drs.derI') 1),
    (auto_tac (cesimps [extend_def])) ]) ;

val _ = bind_thm ("wk_adm_antec_iextrs'", 
  meta_std (rewrite_rule [wk_adm_antec_def] wk_adm_antec_iextrs)) ;
val _ = bind_thm ("wk_adm_antec_iextrs''", 
  meta_std (rewrite_rule [wk_adm_antec_alt] wk_adm_antec_iextrs)) ;

val _ = bind_thm ("wk_adm_antecD", meta_std (wk_adm_antec_def RS defD1)) ;
val _ = bind_thm ("wk_adm_antecD_alt", meta_std (wk_adm_antec_alt RS defD1)) ;
val _ = bind_thm ("iextrs_sw", wk_adm_antec_iextrs RS wk_adm_antecD_alt) ;

val () = qed_goalw_spec_mp "wk_lid_ax" lctr.thy [wk_adm_antec_alt]
  "wk_adm_antec rls --> ({#A#} |- {#A#}) : derrec rls {} --> \
    \ A :# X --> (X |- {#A#}) : derrec rls {}"
  (fn _ => [ Safe_tac, (etac allE2 1), mp_tac 1,
    (auto_tac (cesimps [ms_le_def])) ]) ;

val wk_lid_ax' = drl.singleI RS (drr_empty RS defD2) RSN (2, wk_lid_ax) ;
val _ = bind_thm ("iextrs_lid_ax",
  [wk_adm_antec_iextrs, in_iextrs] MRS wk_lid_ax') ;
val _ = bind_thm ("iextrs_lid_ax'", 
  subset_refl RS reop rev ([idrls.I RS iextrs_lid_ax] MRS der_mono)) ;

Addsimps [symmetric Zero_multiset_def] ;

(** admissibility of left contraction **)

val _ = qed_goalw_spec_mp "l_ax_step" lctr.thy seq_defs
  "({#A#} |- {#}) <= As + As + D --> ({#A#} |- {#}) <= As + D"
  (fn _ => [ (simp_tac (esimps [single_le, ms_atomic]) 1)]) ;

val _ = qed_goal_spec_mp "l_ax_step2" lctr.thy 
  "As + As <= ({#A#} |- {#}) + D --> As <= D"
  (fn _ => [ (rtac impI 1), (dtac le_add_diff_eq 1),
    (ftac (sym RS le_plus_eq1 RS l_ax_step) 1),
    (datac (order_eq_refl RS le_sub_le) 1 1),
    Full_simp_tac 1 ]) ; 

val _ = qed_goal_spec_mp "size_eq_Suc_imp_eq_union'" MS.thy
  "size M = Suc n --> (EX a N. M = N + {#a#} & size N = n)"
  (fn _ => [ Clarify_tac 1, (ftac size_eq_Suc_imp_eq_union 1), 
    Clarsimp_tac 1, (REPEAT (resolve_tac [exI, conjI, refl] 1)) ]) ;

(* getting from contraction of a single formula to contraction of multiple *)
val multi_ctr_antec = refl RS meta_std (prove_goal lctr.thy 
  "ALL A D. (size A = n) --> \
    \ (ALL y: set_of A. ctr_adm_antec rules y) --> \
    \ (A + A |- 0) + D : derrec rules {} --> (A |- 0) + D : derrec rules {}"
  (fn _ => [ (induct_tac "n" 1), Simp_tac 1, Safe_tac,
    (dtac size_eq_Suc_imp_eq_union' 1), Safe_tac, 
    (asm_full_simp_tac (esimps [ball_Un]) 1), Safe_tac,
    (etac allE 1), (etac impE 1), (rtac refl 1), (mp_tac 1), 
    (thin_tac "Ball ?S ?P" 1), 
    (eres_inst_tac [("x", "({#a#} |- 0) + D")] allE 1),
    (etac impE 1), (full_simp_tac (esimps [add_assoc RS sym]) 2),
    (rewtac ctr_adm_antec_def), (etac allE 1), mp_tac 1,
    (dtac mp 1), (simp_tac spsimps 1), 
    (rtac (le_plus2 RS xt6) 1), (simp_tac spsimps 1), 
    (chop_last (etac mem_same 1)),
    (stac (le_pm_assocc RS sym) 1), (ALLGOALS (simp_tac spsimps)) ])) ;

val _ = qed_goal_spec_mp "ctr_antec_alt" lctr.thy
  "Ball ?S (ctr_adm_antec rules) --> (A |- 0) <= G --> \
    \ ((A |- 0) + G : derrec rules {}) --> (G : derrec rules {})"
  (fn _ => [ Safe_tac, (ftac le_add_diff_eq 1),
    (dres_inst_tac [("A", "A"), ("D", "G - (A |- 0)")]
      (multi_ctr_antec) 1), 
    (chop_last (etac rev_mp 2)),
    (stac le_pm_assoc 2), atac 2, (simp_tac spsimps 2),
    (stac le_pm_assoc 1), atac 1,
    (stac (le_pm_assocc RS sym) 1), 
    (ALLGOALS (simp_tac spsimps)),
    (full_simp_tac (esimps plus_ac) 1) ]) ;

val lca1 = [Ball_All, le_plus1] MRS ctr_antec_alt ;
val lca2 = [Ball_All, order_refl, lca1] MRS ctr_antec_alt ;
val _ = bind_thm ("ctr_antec_alt3", reop tl lca2) ;

val _ = qed_goal_spec_mp "ctr_antec_alt2" lctr.thy
  "Ball ?S (ctr_adm_antec rules) --> ((H |- S) : derrec rules {}) --> \
    \ G <= H --> H <= G + G --> ((G |- S) : derrec rules {})"
  (fn _ => [ Safe_tac, 
    (res_inst_tac [("A", "H-G")] ctr_antec_alt 1), atac 1,
    (auto_tac (cesimps [pm_le_eq]))]) ;

val _ = qed_goal_spec_mp "double_adm_antec" lctr.thy
  "Ball (set_of A) (ctr_adm_antec rules)  --> \
    \ (A + A |- S) : derrec rules {} --> (A |- S) : derrec ?rules {}"
  (fn _ => [ Clarify_tac 1, (etac ctr_antec_alt 1), ALLGOALS Simp_tac ]) ;

val _ = bind_thm ("double_eqv_antec", 
  tacsthm [(rtac double_adm_antec 1), atac 2, Simp_tac 2,
    (rtac wk_adm_antec_alt' 2), atac 3, Simp_tac 3,
    (ALLGOALS (etac thin_rl)) ] iffI) ;

val double_eqv_antec' =
  tacsthm [rtac ballI 1, etac thin_rl 1] double_eqv_antec ;

val _ = qed_goal_spec_mp "ctr_wk_le_double" lctr.thy
  "Ball ?S (ctr_adm_antec rls) --> wk_adm_antec rls --> \
    \ (P |- S) : derrec rls {} --> P <= C + C --> (C |- S) : derrec rls {}"
  (fn _ => [ Clarify_tac 1, (eatac (double_eqv_antec RS iffD1) 1 1),
    (eatac wk_adm_antecD_alt 2 1) ]) ;

(* OK TO HERE 
need to look at how much of the rest we need
see which parts of the following are required by ipldt

val _ = qed_goal_spec_mp "wkrls_derl" lctr.thy 
  "extrs {([0], c)} <= derl (extrs wkrls)"
  (fn _ => [ (safe_tac (ces [extrs_E])), 
    (case_tac "c" 1), (clarsimp_tac (cesimps [seq_zero_def]) 1), 
    (res_inst_tac [("M", "multiset1")] multiset_induct 1),
    (res_inst_tac [("M", "multiset2")] multiset_induct 1), Simp_tac 1,
    (ALLGOALS (etac (drl.singleton RS rotate_prems 1 drl.dtderI))),
    (ALLGOALS (rtac extrs.I)), (rtac wkrls.L 3), (rtac wkrls.R 1),
    (ALLGOALS (clarsimp_tac (cesimps [extend_def]))), 
    (REPEAT_FIRST (resolve_tac [conjI, refl])),
    (ALLGOALS (simp_tac spsimps)) ]) ;
    
val _ = qed_goal_spec_mp "gen_wk_tree" lctr.thy
  "extrs wkrls <= rls --> x <= y --> ([x], y) : derl rls"
  (fn _ => [ (rtac impI 1), (rtac impI 1),
    (rtac (singletonI RS extrs.I RS (wkrls_derl RS subsetD) RS drl_mono) 1),
    atac 2, (simp_tac (esimps [extend_def]) 1),
    (rtac conjI 1), (rtac refl 1), (etac le_add_diff_eqc 1) ]) ;

val _ = qed_goal_spec_mp "wk_lengths" lctr.thy 
  "rl : extrs wkrls --> length (fst rl) = 1"
  (fn _ => [ (safe_tac (ces [extrs_E, wkrls.elims])), ALLGOALS Simp_tac ]) ;

val [wk_lengths_derl, wk_lengths_dersl] = map (tacthm (etac wk_lengths 1))
  ([ballI] RL [derl_lengths, dersl_lengths]) ; 

val _ = qed_goal_spec_mp "ctr_lengths" lctr.thy 
  "rl : extrs (ctrrls A) --> length (fst rl) = 1"
  (fn _ => [ (safe_tac (ces [extrs_E, ctrrls.elims])), ALLGOALS Simp_tac ]) ;

val [ctr_lengths_derl, ctr_lengths_dersl] = map (tacthm (etac ctr_lengths 1))
  ([ballI] RL [derl_lengths, dersl_lengths]) ; 

val _ = qed_goal_spec_mp "all_ctr_lengths" lctr.thy 
  "rl : (UN A. extrs (ctrrls A)) --> length (fst rl) = 1"
  (fn _ => [ Safe_tac, (etac ctr_lengths 1) ]) ;

val [all_ctr_lengths_derl, all_ctr_lengths_dersl] = 
  map (tacthm (etac all_ctr_lengths 1))
    ([ballI] RL [derl_lengths, dersl_lengths]) ; 

val _ = bind_thm ("wc_lengths", tacsthm [ etac wk_lengths 2, 
  etac all_ctr_lengths 2 ] UnE) ;

val [wc_lengths_derl, wc_lengths_dersl] = 
  map (tacthm (etac wc_lengths 1)) ([ballI] RL [derl_lengths, dersl_lengths]) ; 

val _ = qed_goal_spec_mp "extrs_wkrls_props" lctr.thy
  "(ps, c) : extrs wkrls --> length ps = 1 & hd ps < c"
  (fn _ => [ Safe_tac, (dtac wk_lengths 1), Force_tac 1,
    (safe_tac (ces [extrs_E, wkrls.elims])),
    (ALLGOALS (clarsimp_tac (claset (), esimps [extend_def, seq_zero_def]
      addsimprocs pmgsps))) ]) ; 

val drs1 = read_instantiate_sg lctr.thy [("P2.0", 
  "%pss cs. (pss, cs) : allrel {(p, c). ([p], c) : (UN x. extrs {([0], x)})}")] 
  drl.induct ;
val drs2 = rotate_prems ~1 (drs1 RS conjunct1 RS mp) ;
val drl_inductsr = tacthm (Force_tac 4) drs2 ;

val _ = qed_goal_spec_mp "wkrls_derl_eq" lctr.thy 
  "derl (extrs wkrls) = (UN c. extrs {([0], c)})"
  (fn _ => [ (rtac equalityI 1), Safe_tac, 
    (etac (mk_mono' wkrls_derl) 2), 
    (etac drl_inductsr 1), (ftac wk_lengths 2),
    (TRYALL (ftac wk_lengths_derl)),
    (TRYALL (ftac wk_lengths_dersl)),
    (ALLGOALS (clarsimp_tac (cesimps [length_Suc_conv]))),
    (rtac exI 1), (rtac (singletonI RS extrs.I) 1),
    (force_tac (cis [add_0], esimps [extend_def]) 1),
    (etac exI 2), (safe_tac (ces [extrs_E, wkrls.elims])), 
    (ALLGOALS (clarsimp_tac (cesimps [extend_def]))),
    (ALLGOALS (rtac exI THEN' rtac (singletonI RS extrs.I))),
    (ALLGOALS (clarsimp_tac (cesimps [extend_def]))),
    Safe_tac, (REPEAT_FIRST (rtac refl)), (ALLGOALS (rtac add_assoc)) ]) ;

(* note also the theorem wk_lengths_derl *)
val _ = qed_goalw "wkrls_derl_iff" lctr.thy [mk_eq wkrls_derl_eq]
  "(([XY], XY') : derl (extrs wkrls)) = (XY <= XY')"
  (fn _ => [ (safe_tac (ces [extrs_E])), 
    (force_tac (cesimps [extend_def]) 1),
    (rtac UN_I 1), (rtac extrs.I 2), TRYALL Fast_tac, 
    (auto_tac (cesimps [extend_def])), (etac le_add_diff_eqc 1) ]) ;

val _ = bind_thm ("wk_admD_alt_ps", 
  rotate_prems 1 (wkrls_derl_iff RS iffD2 RS drl_mono RS derrec_ext)) ;

val _ = qed_goalw "wk_adm_rl" lctr.thy [wk_adm_def, mk_eq adm.unfold]
  "wk_adm rls = (ALL c. extrs {([0], c)} <= adm rls)"
  (fn _ => [ (safe_tac (ces [extrs_E])), 
    (force_tac (cesimps [extend_def']) 1), 
    (etac allE 1), (dtac subsetD 1), (rtac (singletonI RS extrs_I) 1),
    (force_tac (cesimps [drs.all, extend_def']) 1) ]) ;
  
val _ = qed_goalw_spec_mp "wkrls_adm" lctr.thy [mk_eq wk_adm_rl]
  "extrs wkrls <= adm rls --> wk_adm rls"
  (fn _ => [ (Clarify_tac 1), (dtac (mk_mono drl_mono) 1),
    (full_simp_tac (esimps [derl_adm]) 1),
    (etac subsetD 1), (etac (wkrls_derl RS subsetD) 1) ]) ;

val _ = qed_goal "wkrls_adm_iff" lctr.thy 
  "wk_adm rls = (extrs wkrls <= adm rls)"
  (fn _ => [ (rtac iffI 1), (etac wkrls_adm 2),
    (safe_tac (ces [extrs_E] 
      addSDs [wk_admD, wk_alt RS iffD1] addSIs [adm.I])),
    (ALLGOALS Clarsimp_tac), atac 1, 
    (full_simp_tac (esimps [extend_def']) 1) ]) ;

(* contraction rules give contraction admissibility *)
val _ = qed_goalw_spec_mp "ctrrls_adm" lctr.thy [ctr_adm_def, mk_eq adm.unfold] 
  "extrs (ctrrls A) <= adm rls --> ctr_adm rls A"
  (fn _ => [ (safe_tac (cds [ms_of_single])), (ALLGOALS Full_simp_tac), 
    (ALLGOALS (dtac subsetD)), (TRYALL (ares_tac [extrs.I])),
    (TRYALL (ares_tac [refl])), (rtac ctrrls.R 3), (rtac ctrrls.L 1), 
    (ALLGOALS (full_simp_tac (esimps [extend_def]))), Safe_tac, 
    (TRYALL (EVERY' [(etac notE), (etac mem_same),
      (etac (le_add_diff_eq RS sym)) ] )),
    (ALLGOALS (asm_full_simp_tac 
      (esimps [le_pm_assoc, le_mm_assocc RS sym]))) ]) ;

val _ = bind_thm ("wkrls_adm'", (in_adm RSN (2, subset_trans)) RS wkrls_adm) ;
val _ = bind_thm ("ctrrls_adm'", (in_adm RSN (2, subset_trans)) RS ctrrls_adm) ;

val _ = bind_thm ("extrs_wkrls_adm", extrs_mono RS wkrls_adm') ;

(* effect of repeated contractions *)
fun rctacs ctrcon _ = [ (rtac impI 1), (induct_tac "n" 1), (Simp_tac 1),
  (Clarsimp_tac 1), (case_tac "n" 1), (etac mem_same 1), (Simp_tac 1),
  (rtac ms_eqI 1), (force_tac (cesimps [count_mins, count_ms_delete]) 1),
  (rewtac ctr_adm_def), (eres_inst_tac [("x", ctrcon)] allE 1),
  (etac impE 1), (Simp_tac 1), (etac allE 1 THEN mp_tac 1),
  (etac impE 1), (Simp_tac 1), (rtac ms_leI 1), (Force_tac 1),

  (Full_simp_tac 1), (etac allE 1), (etac impE 1),
  (chop_tac 1 (mp_tac 2)), (Asm_simp_tac 1),
  (etac impE 1), (Asm_simp_tac 1),
  (chop_tac 2 (etac mem_same 1)), (rewtac mins_def), (Simp_tac 1),
  (rtac ms_eqI 1), (force_tac (cesimps [count_mins, count_ms_delete]) 1) ] ;

val _ = qed_goal_spec_mp "rep_ctrl'" lctr.thy
  "ctr_adm rls A --> (ALL X. count X A = n --> \
    \ (X |- Y) : derrec rls {} --> A :# X --> \
    \ (mins A (ms_delete {A} X) |- Y) : derrec rls {})" 
  (rctacs "{#A#} |- {#}") ;

val _ = qed_goal_spec_mp "rep_ctrr'" lctr.thy
  "ctr_adm rls A --> (ALL Y. count Y A = n --> \
    \ (X |- Y) : derrec rls {} --> A :# Y --> \
    \ (X |- mins A (ms_delete {A} Y)) : derrec rls {})" 
  (rctacs "{#} |- {#A#}") ;

fun rcutacs str _ = [ Safe_tac, (case_tac str 1), 
  (eresolve_tac [rep_ctrl', rep_ctrr'] 1), (TRYALL (ares_tac [refl])),
  (eatac wk_admD_alt 1 1), Simp_tac 1, (rtac ms_leI 1),
  (force_tac (cesimps [count_mins, count_ms_delete]) 1) ] ;

val _ = qed_goal_spec_mp "rep_ctrl" lctr.thy
  "ctr_adm rls A --> wk_adm rls --> (X |- Y) : derrec rls {} --> \
    \ (mins A (ms_delete {A} X) |- Y) : derrec rls {}" (rcutacs "A :# X") ;

val _ = qed_goal_spec_mp "rep_ctrr" lctr.thy
  "ctr_adm rls A --> wk_adm rls --> (X |- Y) : derrec rls {} --> \
    \ (X |- mins A (ms_delete {A} Y)) : derrec rls {}" (rcutacs "A :# Y") ;

fun sctr_tacs _ = [ (rtac impI 1), (rtac impI 1), (rtac impI 1), 
  (eres_inst_tac [("F", "S")] finite_induct 1), Simp_tac 1,
  Clarsimp_tac 1, (dtac spec 1), (etac thin_rl 1),
  (dresolve_tac ([refl] RLN (2, [rep_ctrl', rep_ctrr'])) 1), atac 1,
  (force_tac (cesimps [count_ms_delete]) 1),
  (chop_tac 1 (etac mem_same 1)),
  (asm_simp_tac (esimps [mins_def, ms_delete_sum,
    singles_insert, ms_delete_singles, union_assoc]) 1) ] ;

val singles_ctrr' = prove_goal lctr.thy
  "All (ctr_adm rls) --> (X |- Y) : derrec rls {} --> finite S --> \
    \ S <= set_of Y --> (X |- singles S + ms_delete S Y) : derrec rls {}" 
  sctr_tacs ;

val singles_ctrl' = prove_goal lctr.thy
  "All (ctr_adm rls) --> (X |- Y) : derrec rls {} --> finite S --> \
    \ S <= set_of X --> (singles S + ms_delete S X |- Y) : derrec rls {}" 
  sctr_tacs ;

val singles_ctrr = simplify (esimps [])
  ([asm_rl, asm_rl, finite_set_of, order_refl] MRS meta_std singles_ctrr') ;
val singles_ctrl = simplify (esimps [])
  ([asm_rl, asm_rl, finite_set_of, order_refl] MRS meta_std singles_ctrl') ;

val _ = qed_goal_spec_mp "singles_ctr" lctr.thy
  "All (ctr_adm rls) --> XY : derrec rls {} --> seq_singles XY : derrec rls {}"
  (fn _ => [ (case_tac "XY" 1), Clarsimp_tac 1,
    (fatac singles_ctrr 1 1), (eatac singles_ctrl 1 1) ]) ;

val _ = qed_goal_spec_mp "singles_ctr_derl'" lctr.thy
  "XY = (X |- Y) --> S = (Sl |- Sr) --> XY - S >= seq_singles XY --> \
    \ ([XY], XY - S) : derl (UN A. extrs (ctrrls A))"
  (fn _ => [ (rtac impI 1), (rtac impI 1), hyp_subst_tac 1,
    (res_inst_tac [("M", "Sl")] multiset_induct 1),
    (res_inst_tac [("M", "Sr")] multiset_induct 1), Simp_tac 1,
    (ALLGOALS (EVERY' [(rtac impI), (etac impE), (etac order_trans),
      (rtac sub_le), Simp_tac ])),

    (case_tac "x :# X" 2), (case_tac "x :# Y" 1),
    let val mstacs = EVERY' [ (etac mem_same),
      (simp_tac (esimps [pm0_class.pm0_minus_minus RS sym])),
      (rtac ms_eqI), Force_tac ] in mstacs 4 THEN mstacs 2 end,
    
    (ALLGOALS (etac (drl.singleton RS rotate_prems 1 drl.dtderI))),
    (ALLGOALS (res_inst_tac [("a", "x")] (UNIV_I RS UN_I))),
    (REPEAT_FIRST (resolve_tac [UNIV_I RS UN_I, extrs.I])),

    (rtac ctrrls.L 3), (rtac ctrrls.R 1), 
    (ALLGOALS (full_simp_tac (esimps [extend_def]))), Safe_tac, 

    (TRYALL (rtac le_add_diff_eq)), (ALLGOALS (simp_tac spsimps)),
    (TRYALL (etac (rotate_prems 1 order_trans) THEN'
      asm_simp_tac (esimps [single_le]))),
    (ALLGOALS (resolve_tac [ms_eqI])), (ALLGOALS (case_tac "x = xa")),
    rewtac ms_le_def, (REPEAT_FIRST (eres_inst_tac [("x", "xa")] allE)),
    (ALLGOALS (clarsimp_tac (cesimps [count_singles, singles_count_def]))),
    ALLGOALS arith_tac ]) ;

val singles_ctr_derl = 
  [antec_succ_simp RS sym, antec_succ_simp RS sym] MRS singles_ctr_derl' ;

val drc1 = read_instantiate_sg lctr.thy [
  ("P1.0", "%ps c. (ps, [c]) : allrel ?R"), 
  ("P2.0", "%pss cs. (pss, cs) : allrel ?R")] drl.induct ;
val drc2 = rotate_prems ~1 (drc1 RS conjunct1 RS mp) ;
val drl_induct_ctr = tacthm (Force_tac 4) drc2 ;

val _ = qed_goal_spec_mp "ctrrls_props" lctr.thy
  "(ps, c) : ctrrls A --> \
    \ length ps = 1 & hd ps > c & seq_singles (hd ps) = seq_singles c"
  (fn _ => [ (auto_tac (ces [ctrrls.elims], esimps [] addsimprocs pmgsps)) ]) ;

val _ = qed_goal_spec_mp "extrs_ctrrls_props" lctr.thy
  "(ps, c) : extrs (ctrrls A) --> \
    \ length ps = 1 & hd ps > c & seq_singles (hd ps) = seq_singles c"
  (fn _ => [ (auto_tac (ces [extrs_E, make_elim ctrrls_props], esimps [])),
    (ALLGOALS (case_tac "aa")), (auto_tac (cesimps [extend_def])),
    (case_tac "flr" 1), (case_tac "a" 1), (case_tac "ba" 1),
    (auto_tac (cesimps [extend_def])) ]) ;

(* note also the theorem ctr_lengths_derl *)
val _ = qed_goal_spec_mp "ctrrls_derl_iff" lctr.thy
  "(([XY], XY') : derl (UN A. extrs (ctrrls A))) = \
    \ (XY' <= XY & XY' >= seq_singles XY)"
  (fn _ => [ Safe_tac, (dtac le_minus_diff 3),
    (fatac xt3 1 3), (dtac singles_ctr_derl 3), Force_tac 3,
    (ALLGOALS (dtac drl_induct_ctr)),
    (REPEAT_FIRST (eresolve_tac 
      [in_Collect_Pair_iff RS iffD1, allrel_ConsE])),
    (safe_tac (ces [allrelE1', allrelE2'])), (ALLGOALS Clarsimp_tac), 
    (ALLGOALS (dtac extrs_ctrrls_props)), (ALLGOALS (case_tac "ps")), 
    (ALLGOALS (clarsimp_tac (ces [allrelE1', allrelE2'], esimps []))), 
    (eatac (xt8 RS order_less_imp_le) 1 1), (dtac seq_singles_mono 1),
    (rtac (seq_singles_le RS xt6) 1), Clarsimp_tac 1 ]) ;

val _ = bind_thm ("singles_ctrrls",
    [seq_singles_le, order_refl] MRS ([ctrrls_derl_iff, conjI] MRS iffD2)) ;

val _ = qed_goal "ctr_adm_rls" lctr.thy
  "ctr_adm rls A = (extrs (ctrrls A) <= adm rls)"
  (fn _ => [ (rtac iffI 1), (etac ctrrls_adm 2), 
    (clarsimp_tac (ces [extrs_E], esimps [ctr_adm_def]) 1),
    (safe_tac (ces [ctrrls.elims] addSIs [adm.I])),
    (ALLGOALS Clarsimp_tac),
    (eres_inst_tac [("x", "({#A#} |- 0)")] allE 1),
    (eres_inst_tac [("x", "(0 |- {#A#})")] allE 2),
    (REPEAT_FIRST (eresolve_tac [allE, impE])), (TRYALL atac), 
    (ALLGOALS (full_simp_tac (esimps [extend_def]))), 
    (ALLGOALS (chop_tac 1 o etac mem_same)),
    (ALLGOALS (simp_tac (esimps [le_pm_assocc RS sym]))) ]) ;

val _ = qed_goal "derl_ctr_adm" lctr.thy
  "All (ctr_adm rls) = (derl (UN A. extrs (ctrrls A)) <= adm rls)"
  (fn _ => [ (rtac iffI 1), (rtac xt3 1),
    (rtac (derl_adm RS sym) 1), (rtac (mk_mono drl_mono) 1),
    Safe_tac, (dtac spec 1),
    (dtac (ctr_adm_rls RS iffD1) 1), (eatac subsetD 1 1),
    (rtac (ctr_adm_rls RS iffD2) 1), 
    (dtac (drl.in_derl RS subset_trans) 1), Fast_tac 1 ]) ;

val _ = bind_thm ("ctr_singles_adm", 
  [derl_ctr_adm RS iffD1 , singles_ctr_derl] MRS subsetD) ;

val _ = bind_thm ("singles_ctr'", 
  full_simplify (esimps [mk_eq drs.all]) (ctr_singles_adm RS admD)) ;

val _ = qed_goal_spec_mp "singles_ctrd" lctr.thy
  "All (ctr_adm rls) --> XY : derrec rls {} --> \
    \ XY' <= XY --> XY' >= seq_singles XY --> XY' : derrec rls {}"
  (fn _ => [ Clarify_tac 1, (dtac singles_ctr' 1), (TRYALL atac),
    (etac mem_same 2 THEN etac le_minus_diff 2), Clarsimp_tac 1 ]) ;

val ctd = [ctrrls_derl_iff, conjI] MRS iffD2 RS drl_mono ;

val _ = qed_goal_spec_mp "singles_ctrd_ps" lctr.thy
  "(UN A. extrs (ctrrls A)) <= rls --> XY : derrec rls prems --> \
    \ XY' <= XY --> XY' >= seq_singles XY --> XY' : derrec rls prems"
  (fn _ => [ Safe_tac, (datac ctd 2 1),
    (etac (singleton_subsetI RS derrec_trans) 1),
    (etac (fold_rule [derl_rr] derrc.rcI) 1), Simp_tac 1 ]) ;

val mms = transfer lctr.thy singles_le RS le_minus_diff ;

val _ = bind_thm ("singles_ctr", 
  [asm_rl, asm_rl, seq_singles_le, order_refl] MRS singles_ctrd) ;
val _ = bind_thm ("singles_ctr_ps", 
  [asm_rl, asm_rl, seq_singles_le, order_refl] MRS singles_ctrd_ps) ;

(* could also get the next few in terms of 
  derl (wk_adm rls Un (UN A. ctr_adm rls A)),
  just like ctrrls_derl_iff and derl_ctr_adm *)

(* note also the theorem wc_lengths *)
val _ = qed_goal "wk_ctr_derl_iff" lctr.thy 
  "(([XY], XY') : derl (extrs wkrls Un (UN A. extrs (ctrrls A)))) = \
    \ (seq_singles XY' >= seq_singles XY)"
  (fn _ => [ rtac iffI 1,
    (dtac (seq_singles_le RS xt6) 2), (dtac (wkrls_derl_iff RS iffD2) 2),
    (dtac drl_mono 2), (rtac Un_upper1 2), (etac derl_trans 2), Simp_tac 2, 
    (rtac (singles_ctrrls RS drl_mono) 2), (rtac Un_upper2 2),

    (dtac drl_induct_ctr 1),
    (REPEAT_FIRST (eresolve_tac 
      [in_Collect_Pair_iff RS iffD1, allrel_ConsE])),
    (safe_tac (ces [allrelE1', allrelE2'])), (ALLGOALS Clarsimp_tac), 
    (dtac extrs_ctrrls_props 2), (dtac extrs_wkrls_props 1),
    (ALLGOALS (case_tac "ps")), 
    (ALLGOALS (clarsimp_tac (ces [allrelE1', allrelE2'], esimps []))), 
    (dtac (order_less_imp_le RS seq_singles_mono) 1),
    (eatac order_trans 1 1) ]) ;

val lea_dlea = [derl_adm RS sym, mk_mono drl_mono] MRS xt3 ;

val _ = qed_goal_spec_mp "wk_ctr_adm_singles_iff" lctr.thy 
  "(wk_adm rls & All (ctr_adm rls)) = \
    \ (ALL s c. seq_singles c >= seq_singles s --> \
      \ s : derrec rls {} --> c : derrec rls {})"
  (fn _ => [ rtac iffI 1, Clarify_tac 1, (rtac admD 1), 
    (etac drs.dlCons 2), (rtac drs.dlNil 2),
    (dtac (wk_ctr_derl_iff RS iffD2) 1),
    (etac (mk_mono' lea_dlea) 1),
    SELECT_GOAL Safe_tac 1, (force_tac (cesimps [wkrls_adm_iff]) 1),
    (force_tac (cesimps [ctr_adm_rls]) 1),
    (subgoal_tac "extrs wkrls Un (UN A. extrs (ctrrls A)) <= adm rls" 1),
    (rtac (drl.in_derl RS order_trans) 2),
    (Clarify_tac 2), (ftac wc_lengths_derl 2),
    (case_tac "a" 2), (ALLGOALS Clarsimp_tac),
    (dtac (wk_ctr_derl_iff RS iffD1) 2),
    (etac allE2 2), mp_tac 2, (rtac adm.I 2), Force_tac 2, 
    (fast_tac (cis [wkrls_adm, ctrrls_adm]) 1) ]) ;

val _ = bind_thm ("wk_ctr_adm_singles",
  meta_std (wk_ctr_adm_singles_iff RS iffD1)) ;

(* proof using wk_ctr_adm_singles of wk_ctr_adm_setof,
  not really useful, just to show we can *)
val _ = qed_goal_spec_mp "wk_ctr_adm_setof" lctr.thy 
  "(wk_adm rls & All (ctr_adm rls)) = \
    \ (ALL Xp Xc Yp Yc. set_of Xp <= set_of Xc --> set_of Yp <= set_of Yc --> \
    \ (Xp |- Yp) : derrec rls {} --> (Xc |- Yc) : derrec rls {})"
  (fn _ => [ (rtac (wk_ctr_adm_singles_iff RS trans) 1), Safe_tac,
    (etac allE2 1), (etac impE 1), (eatac mp 1 2), Asm_simp_tac 1,
    (case_tac "s" 1), (case_tac "c" 1), Clarsimp_tac 1 ]) ; 

val _ = qed_goal_spec_mp "wk_ctr_adm_del" lctr.thy 
  "wk_adm rls --> All (ctr_adm rls) --> s : derrec rls {} --> \
    \ seq_delete p s + p : derrec rls {}"
  (fn _ => [ (case_tac "s" 1), (case_tac "p" 1), Clarsimp_tac 1, 
    (eatac (meta_std (wk_ctr_adm_setof RS iffD1)) 1 1), atac 3, 
    (ALLGOALS (simp_tac (esimps [set_of_delete, Un_upper2]))),
    (TRYALL Fast_tac) ]) ;

(* to use a rule plus weakening and contraction in forward proof *)
val _ = qed_goal_spec_mp "wk_ctr_adm_rule_sub" lctr.thy 
  "wk_adm rls --> All (ctr_adm rls) --> \
    \ extrs {([p], c)} <= adm rls --> s : derrec rls {} --> \
    \ seq_delete p s + c : derrec rls {}"
  (fn _ => [ Safe_tac, (dtac subsetD 1),
    (rtac (singletonI RS extrs_I) 1), (etac adm.elim 1),
    (clarsimp_tac (cesimps [extend_def', drs.all]) 1),
    (etac mp 1), (eatac wk_ctr_adm_del 2 1) ]) ;

val _ = qed_goal "seqid_split" lctr.thy "(A |- A) = (0 |- A) + (A |- 0)" fn_at ;
val _ = qed_goal "seq_split" lctr.thy "(A |- B) = (0 |- B) + (A |- 0)" fn_at ;
val seq_split' = [seq_split, add_commute] MRS trans ;

(* following two very similar except that one involves premises to the
  derivation, and so cannot use weakening admissibility *)
fun gcl_tacs wkth _ = [ (clarsimp_tac (cesimps [gen_step.simps]) 1),
  (etac extrs.elims 1), (subgoal_tac "etr M dseq b = c + etr M dseq flr" 1),
  (force_tac (claset (), esimps [extend_def, etr_def,
    seq_delete_sum, seq_delete_disj, pg_disj_minus] 
    addsimprocs pmgsps) 2),
  Clarsimp_tac 1, (rtac drs.derI 1),
  (dtac (singletonI RS extrs_I RS rev_subsetD) 1),
  (etac mem_same 1), (force_tac (cesimps [extend_def]) 1),
  (clarsimp_tac (cesimps [drs.all]) 1), (datac bspec 1 1),
  (etac wkth 1), (etac conjunct2 1),
  (simp_tac (esimps [extend_def, etr_def,
    seq_delete_sum] addsimprocs pmgsps) 1) ] ;

val _ = qed_goalw_spec_mp "gen_comm_lemg_ps" lctr.thy [gsfg_ps_def]
  "extrs wkrls <= prls --> extrs {(ps, c)} <= prls --> pg_meet c dseq = 0 --> \
    \ Ball (extrs {(ps, c)}) (gsfg_ps prems prls drls (etr M dseq) A)" 
  (gcl_tacs wk_admD_alt_ps) ;

val _ = qed_goalw_spec_mp "gen_comm_lemg" lctr.thy [gsfg_def]
  "wk_adm prls --> extrs {(ps, c)} <= prls --> pg_meet c dseq = 0 --> \
    \ Ball (extrs {(ps, c)}) (gsfg prls drls (etr M dseq) A)" 
  (gcl_tacs wk_admD_alt) ;

val _ = qed_goalw "gsf_gsfg" lctr.thy [meta_eq_def]
  "gsf rls == gsfg rls rls" 
  (fn _ => [ (clarsimp_tac (cis [ext], esimps [gsf_def, gsfg_def]) 1) ]) ;

val _ = qed_goalw "gsfg_ps" lctr.thy [meta_eq_def]
  "gsfg == gsfg_ps {}" 
  (fn _ => [ (clarsimp_tac (cis [ext], esimps [gsfg_ps_def, gsfg_def]) 1) ]) ;

val gsf_eq_gsfg = gsf_gsfg RS def_imp_eq RS fun_cong RS fun_cong RS fun_cong ;
fun mk_gsf gth = [gsf_eq_gsfg, gth] MRS iffD2 ;

val _ = qed_goalw_spec_mp "gen_comm_lem" lctr.thy [gsf_gsfg]
  "wk_adm rls --> extrs {(ps, c)} <= rls --> pg_meet c dseq = 0 --> \
    \ Ball (extrs {(ps, c)}) (gsf rls (etr M dseq) A)" 
  (fn _ => [ (safe_tac (cis [gen_comm_lemg])), TRYALL Fast_tac ]) ;

(* following two very similar except that one involves premises to the
  derivation, and so cannot use weakening admissibility *)
fun wel_tacs wkth _ = [ (clarsimp_tac (cesimps [gen_step.simps]) 1),
  (etac extrs.elims 1), (etac wkrls.elims 1),
  (ALLGOALS Clarsimp_tac),
  (ALLGOALS (EVERY' [ (dtac asm_rl), (etac thin_rl), (etac thin_rl),
    (eatac wkth 1), (rtac etr_mono)])),
  (TRYALL (rtac order_refl)),
  (ALLGOALS (simp_tac (esimps [extend_def]))) ] ;

val _ = qed_goalw_spec_mp "wk_extrs_lemg_ps" lctr.thy [gsfg_ps_def]
  "extrs wkrls <= rls --> (ps, c) : wkrls --> Ball (extrs {(ps, c)}) \
    \ (gsfg_ps prems rls drls (etr M dseq) A)" 
  (wel_tacs wk_admD_alt_ps) ;

val _ = qed_goalw_spec_mp "wk_extrs_lemg" lctr.thy [gsfg_def]
  "wk_adm rls --> (ps, c) : wkrls --> \
    \ Ball (extrs {(ps, c)}) (gsfg rls drls (etr M dseq) A)" 
  (wel_tacs wk_admD_alt) ;

val rsc = rotate_prems 1 singles_ctrd ;
val rsc_ps = rotate_prems 1 singles_ctrd_ps ;

fun ctl_tacs rsc_th _ = [ (clarsimp_tac (cesimps [gen_step.simps]) 1),
  (etac extrs.elims 1), (etac ctrrls.elims 1),
  (ALLGOALS Clarsimp_tac),
  (ALLGOALS (EVERY' [ (dtac asm_rl), (etac thin_rl), (etac thin_rl)])),
  (ALLGOALS (etac rsc_th)), (TRYALL atac), 
  (TRYALL (EVERY' [rtac etr_mono, REPEAT1 o rtac order_refl, 
    (case_tac "flr"), (simp_tac (esimps [extend_def])) ])),
  (ALLGOALS (EVERY' [ (case_tac "M"), (case_tac "dseq"), (case_tac "flr") ])),
  (ALLGOALS (clarsimp_tac (cesimps [extend_def, etr_def, ms_delete_sum]))),
  (REPEAT_FIRST (resolve_tac [transfer MS.thy singles_Un RS order_trans,
    conjI, finite_set_of, add_mono, singles_le,
    [finite_Un, conjI] MRS iffD2])) ] ;

val _ = qed_goalw_spec_mp "ctr_extrs_lemg_ps" lctr.thy [gsfg_ps_def]
  "(UN A. extrs (ctrrls A)) <= rls --> (ps, c) : ctrrls C --> \
    \ Ball (extrs {(ps, c)}) (gsfg_ps prems rls drls (etr M dseq) A)" 
  (ctl_tacs rsc_ps) ;

val _ = qed_goalw_spec_mp "ctr_extrs_lemg" lctr.thy [gsfg_def]
  "All (ctr_adm rls) --> (ps, c) : ctrrls C --> \
    \ Ball (extrs {(ps, c)}) (gsfg rls drls (etr M dseq) A)" 
  (ctl_tacs rsc) ;

val wk_extrs_lem = mk_gsf wk_extrs_lemg ;
val ctr_extrs_lem = mk_gsf ctr_extrs_lemg ;

(* isolate step in the gen_inv proof *)
val _ = qed_goal_spec_mp "gen_inv_step'" lctr.thy
  "Ball rls (inv_step (derrec rls {}) irls) --> c : derrec rls {} --> \
    \ invs_of irls c <= derrec rls {}"
  (fn _ => [ (REPEAT (rtac impI 1)),
    (etac drs.inductr 1), Fast_tac 1,
    (force_tac (cesimps [inv_step.simps, drs.all]) 1) ]) ;
    
val _ = qed_goal_spec_mp "gen_inv_step" lctr.thy
  "Ball rls (inv_step (derrec rls {}) irls) --> (ps, c) : irls --> \
    \ c : derrec rls {} --> ps : dersrec rls {}"
  (fn _ => [ (REPEAT (rtac impI 1)), (datac gen_inv_step' 1 1),
    (force_tac (cesimps [invs_of_def, drs.all]) 1) ]) ;

val _ = qed_goal_spec_mp "gen_inv_by_step" lctr.thy
  "Ball rls (inv_step (derrec rls {}) irls) --> Ball irls (inv_rl rls)"
  (fn _ => [ Clarify_tac 1, (simp_tac (esimps [inv_rl.simps]) 1),
    (rtac impI 1), (eatac gen_inv_step 2 1) ]) ;
    
(* alternatively, proof using gen_step_lem *)
val _ = qed_goal_spec_mp "gen_inv_by_step" lctr.thy
  "Ball rls (inv_step (derrec rls {}) irls) --> Ball irls (inv_rl rls)"
  (fn _ => [ (clarsimp_tac (cesimps [inv_rl.simps]) 1),
    (dres_inst_tac [("P", "%A x. invs_of irls x <= derrec rls {}")] 
      (reop rev gen_step_lem) 1),
    (rtac wfp_empty 2), (force_tac (cesimps [invs_of_def, drs.all]) 2),
    (force_tac (cesimps [Ball_def, inv_step.simps, gen_step.simps]) 1) ]) ;

val iss_tacs = [ 
  (clarsimp_tac (ces [extrs.elims], esimps [inv_step.simps,
    mod_inv_step.simps, mod_inv_prop_def, invs_of_def]) 1), 
  (thin_tac "Ball ?S ?P" 1),
  (datac (imageI RS rev_subsetD) 1 1),
  (asm_full_simp_tac (esimps [extend_def]) 1) ] ;

val _ = qed_goal_spec_mp "inv_stepm_same" lctr.thy 
  "srl : extrs {rl} --> inv_stepm rls (extrs {rl}) srl"
  (fn _ => [
    (clarsimp_tac (ces [extrs.elims], esimps [inv_stepm.simps, invs_of_def]) 1),
    (rtac drs.dpI 1), (rtac UnI1 1), (etac rev_image_eqI 1),
    (asm_full_simp_tac (esimps [extend_def]) 1) ]) ;

val _ = qed_goal_spec_mp "mod_inv_step_same" lctr.thy 
  "srl : extrs {rl} --> mod_inv_step (derrec rls {}) (extrs {rl}) f srl"
  (fn _ => iss_tacs) ;

val _ = qed_goal_spec_mp "inv_stepm_same_mul" lctr.thy 
  "srl : extrs {rl} --> srl : rls --> \
    \ inv_stepm rls (extrs (multiples {rl})) srl"
  (fn _ => [
    (clarsimp_tac (ces [extrs.elims, multiples.elims],
      esimps [inv_stepm.simps, invs_of_def]) 1),
    (case_tac "n" 1), Clarsimp_tac 1,
    (etac drs.derI' 1), Clarsimp_tac 1, 
    (rtac drs.dpI 1), (rtac UnI1 1), (etac imageI 1), 

    (clarsimp_tac (claset (), esimps [extend_def] addsimprocs pmgsps) 1), 
    (rtac drs.dpI 1), (rtac UnI2 1), rtac UN_I 1, atac 1,
    (rtac CollectI 1), rtac exI 1, rtac conjI 1, rtac extrs.I 2,
    (res_inst_tac [("n", "nat")] multiples.I 2), rtac singletonI 2, 
    (clarsimp_tac (claset (), esimps [extend_def] addsimprocs pmgsps) 2), 
    rtac conjI 2, rtac refl 3, rtac refl 2,
    (simp_tac (esimps [image_compose RS sym]) 1), (etac rev_image_eqI 1), 
    (simp_tac (esimps [extend_def] addsimprocs pmgsps) 1) ]) ;

val _ = bind_thm ("inv_step_same", inv_stepm_same RS inv_step_m) ;
val _ = bind_thm ("inv_step_same_mul", inv_stepm_same_mul RS inv_step_m) ;

val _ = qed_goal_spec_mp "mod_inv_id" lctr.thy 
  "idrl = ([], {#A#} |- {#A#}) --> srl : extrs {idrl} --> f A --> \
    \ extrs {idrl} <= rls --> mod_inv_step (derrec rls {}) any f srl"
  (fn _ => [ 
    (clarsimp_tac (ces [extrs.elims], esimps [
      mod_inv_step.simps, mod_inv_prop_def, extend_def,
      seq_filter_sum]) 1), 
    (etac notE 1), (asm_simp_tac (esimps [ms_filter_single]) 1),
    (rtac drs.derINil 1), (etac subsetD 1), 
    (rtac (singletonI RS extrs.I) 1),
    (simp_tac (esimps [extend_def]) 1) ]);

(* inv_step by commuting rules *)
val _ = qed_goalw_spec_mp "commg_inv_stepm" lctr.thy [commg_rules_def]
  "(ALL irl : irls. commg_rules srl irl srls irls) --> inv_stepm srls irls srl"
  (fn _ => [ (case_tac "srl" 1),
    (clarsimp_tac (cesimps [inv_stepm.simps, invs_of_def]) 1), 
    (datac bspec 1 1),
    (etac allE 1 THEN dtac (refl RS rev_mp) 1),
    (etac allE 1 THEN mp_tac 1), Clarify_tac 1, 
    (rtac derrec_trans 1), atac 2, (rtac subsetI 1),
    (datac bspec 1 1), Clarify_tac 1,
    (rtac drs.dpI 1), rtac UnI2 1, Fast_tac 1 ]) ;

val _ = qed_goalw_spec_mp "commg_comm" lctr.thy 
  [comm_rules_def, commg_rules_def]
  "comm_rules srl irl srls irls --> commg_rules srl irl srls irls"
  (fn _ => [ (rtac impI 1), 
    (REPEAT (eresolve_tac [all_forward, imp_forward] 1)),
    Clarify_tac 1, (rewtac derl_rr),
    (dtac ([drl.singleI, subset_refl] MRS derrc.rcI) 1),
    (rtac exI 1), (eatac conjI 1 1) ]) ;

val _ = qed_goal_spec_mp "comm_inv_stepm" lctr.thy 
  "(ALL irl : irls. comm_rules srl irl srls irls) --> inv_stepm srls irls srl"
  (fn _ => [ Safe_tac, (rtac commg_inv_stepm 1),
    (REPEAT (eresolve_tac [ball_match, commg_comm, asm_rl] 1)) ]) ;

val _ = bind_thm ("commg_inv_step", 
  commg_inv_stepm RS inv_step_mono RS inv_step_m) ;
val _ = bind_thm ("comm_inv_step", 
  comm_inv_stepm RS inv_step_mono RS inv_step_m) ;

val _ = qed_goalw_spec_mp "comm_rules_disj" lctr.thy [comm_rules_def]
  "seq_meet c ic = 0 --> extrs {(ps, c)} = srls --> srl : srls --> \
    \ extrs {(ips, ic)} = irls --> irl : irls --> \
    \ comm_rules srl irl srls irls" 
  (fn _ => [ (safe_tac (ces [extrs_E])),
    (clarsimp_tac (cesimps [extend_def]) 1),
    (subgoal_tac "c <= flra" 1), (etac seq_disj_le_add 2), 
    (etac (sym RS nth (xtrans, 2)) 2), (rtac le_plus1 2),
    (dtac le_add_diff_ex 1), (Clarify_tac 1), 
    (res_inst_tac [("x", "map (extend (x + D)) ps")] exI 1),
    Safe_tac, (clarsimp_tac (cesimps [extend_def]) 2), (rtac bexI 2),
    (res_inst_tac [("flr", "D + xa")] (singletonI RS extrs_I) 3),
    (clarsimp_tac (cesimps [extend_def]) 2),
    (rtac conjI 2), (TRYALL (etac rev_image_eqI)),
    (res_inst_tac [("flr", "x + D")] (singletonI RS extrs.I) 1),
    (ALLGOALS (asm_full_simp_tac spsimps)) ]) ;

val _ = qed_goal_spec_mp "inv_stepm_disj_cs" lctr.thy
  "seq_meet c ic = 0 --> extcs {(ps, c)} = rls --> rl : rls --> \
    \ inv_stepm rls (extrs {(ips, ic)}) rl"
  (fn _ => [ 
    (clarsimp_tac (ces [extrs_E, extcs_E],
      esimps [extend_def, inv_stepm.simps, invs_of_def]) 1), 
    (* no need to invert from premises *)
    (rtac ([asm_rl, subset_refl, Un_upper1] MRS der_mono) 1), 
    (subgoal_tac "c <= flra" 1), (etac seq_disj_le_add 2), 
    (etac (nth (xtrans, 2)) 2), (rtac le_plus1 2),
    (dtac le_add_diff_ex 1), (Clarify_tac 1), 
    (rtac drs.derI 1), (rtac (singletonI RS extcs.I) 1),
    (simp_tac spsimps 1), Fast_tac 1, 
    (rtac ([drs.all, drs.sdp] MRS iffD2) 1) ]) ;

val _ = qed_goal_spec_mp "inv_stepm_disj" lctr.thy
  "seq_meet c ic = 0 --> extrs {(ps, c)} = rls --> rl : rls --> \
    \ inv_stepm rls (extrs {(ips, ic)}) rl"
  (fn _ => [ Safe_tac, (rtac comm_inv_stepm 1), (rtac ballI 1),
    (etac comm_rules_disj 1), (TRYALL (ares_tac [refl])) ]) ;

val _ = qed_goal_spec_mp "inv_stepm_disj_times" lctr.thy
  "seq_meet c ic = 0 --> extrs {(ps, c)} = rls --> rl : rls --> \
    \ inv_stepm rls (extrs {pscmap (times n) (ips, ic)}) rl"
  (fn _ => [ Safe_tac, (rtac comm_inv_stepm 1), (rtac ballI 1),
    Full_simp_tac 1, (rtac comm_rules_disj 1), (TRYALL (ares_tac [refl])),
    rewtac seq_meet_pg, (etac meet_times_0' 1) ]) ;

val isdtm = prove_goal lctr.thy 
  "(!n. inv_stepm rls (extrs {pscmap (times n) (ips, ic)}) rl) --> \
    \ inv_stepm rls (extrs (multiples {(ips, ic)})) rl"
  (fn _ => [ (clarsimp_tac (cesimps [multiples_eq, extrs_UN RS sym]) 1), 
    (rtac inv_stepm_UN 1), Asm_simp_tac 1 ]) ;

val _ = qed_goal_spec_mp "inv_stepm_disj_mul" lctr.thy
  "seq_meet c ic = 0 --> extrs {(ps, c)} = rls --> rl : rls --> \
    \ inv_stepm rls (extrs (multiples {(ips, ic)})) rl"
  (fn _ => [ Safe_tac, (rtac (isdtm RS mp) 1), rtac allI 1, 
    (rtac inv_stepm_disj_times 1), (TRYALL (ares_tac [refl])) ]) ;

val inv_stepm_disj' = inv_stepm_disj RS inv_step_mono ;
val _ = bind_thm ("inv_step_disj", inv_stepm_disj' RS inv_step_m) ;
val _ = bind_thm ("inv_step_disj_times", 
  inv_stepm_disj_times RS inv_step_mono RS inv_step_m) ;
val inv_stepm_disj_mul' = inv_stepm_disj_mul RS inv_step_mono ;
val _ = bind_thm ("inv_step_disj_mul", inv_stepm_disj_mul' RS inv_step_m) ;

val _ = qed_goalw_spec_mp "id_decomp_step1" lctr.thy [id_decomp_def]
  "id_decomp rls --> rls' <= rls --> \
    \ inv_step (derrec rls {}) rls' (ps, {#A#} |- {#A#})"
  (fn _ => [ (simp_tac (esimps [inv_step.simps, invs_of_def]) 1),
    Auto_tac ]) ;

val _ = qed_goal_spec_mp "id_decompm_step1e" lctr.thy 
  "id_decomp (extrs rls) --> srl : rls --> srl : scrls --> \
    \ irl : rls --> irl : idrls --> pscmap (extend flr) irl = psc --> \
    \ inv_stepm (extrs rls) (extrs {srl}) psc"
  (fn _ => [ Safe_tac,
    (case_tac "seq_meet (snd irl) (snd srl) = 0" 1),
    (* case where the inversion rule is not in the {#A#} |- {#A#} part *)
    (case_tac "srl" 1), Clarify_tac 1,
    (rtac inv_stepm_disj' 1), Full_simp_tac 1,
    (rtac refl 1), (rtac (singletonI RS extrs.I) 1),
    gen_cong_tac 1, (rtac refl 1), (rtac (thm "pair_collapse") 1),
    (rtac extrs_mono 1), Force_tac 1,
    (* case where the inversion rule is in the {#A#} |- {#A#} part *)
    (etac swap 1), (rtac (mk_seq meet_commute RS trans) 1), (etac swap 1),
    (etac scrls.elims 1), (etac idrls.elims 1),
    (dtac (order_eq_refl RS seq_atom_disj_or_le) 1),
    Clarsimp_tac 1, Safe_tac, contr_tac 2,  (* level 22 *)
    (etac (le_add_diff_ex RS exE) 1), (rewtac id_decomp_def), 
    (eres_inst_tac [("x", "map (extend x) ps")] allE2 1), (etac impE 1),
    (clarsimp_tac (cesimps [inv_stepm.simps, invs_of_def]) 2), 
    (etac extrs.elims 2), Clarsimp_tac 2, (dtac subsetD 2), Fast_tac 2,
    (dres_inst_tac [("As", "flr")] (wk_adm_extrs RS wk_admD) 2),
    (etac mem_same 2), (dtac sym 2), 
    (asm_full_simp_tac spsimps 2), 
    (etac extrs.I 1), (simp_tac (esimps [extend_def]) 1), 
    (rtac conjI 1), atac 2, Fast_tac 1 ]) ;

val id_decompm_step1e' = idrls.I RS 
  reop (fn [a,b,c,d,e,f] => [e,a,f,d,c,b]) id_decompm_step1e ;

val _ = bind_thm ("id_decomp_step1e", id_decompm_step1e RS inv_step_m) ;

(*
val id_decomp_step1e' = idrls.I RS 
  reop (fn [a,b,c,d,e,f] => [e,a,f,d,c,b]) id_decomp_step1e ;
*)
val id_decomp_step1e' = id_decompm_step1e' RS inv_step_m ;
val id_decomp_step1e'' = read_instantiate [("flr2", "?fl |- ?fr")]
  (id_decompm_step1e RS inv_step_mono RS inv_step_m) ; 

val _ = bind_thm ("id_le_sum", 
  full_simplify (esimps []) (seq_atomic_singles MRS atoms_le_sum)) ;
val _ = bind_thm ("id_le_twice", 
  full_simplify (esimps []) (seq_atomic_singles MRS atoms_le_twice)) ;

val _ = qed_goal_spec_mp "id_times_lem" lctr.thy
  "extend flr (times n c + c) >= ({#A#} |- {#A#}) --> \
    \ extend flr c >= ({#A#} |- {#A#})"
  (fn _ => [ (induct_tac "n" 1), 
    (ALLGOALS (full_simp_tac (esimps [extend_def]))),
    rtac impI 1, (etac mp 1), 
    (forw_inst_tac [("c", "times n c + flr"), ("d", "c + c")] 
      (thin_rl RS id_le_sum) 1),
    (etac (reop rev xt3) 1), (simp_tac (esimps add_ac) 1),
    Safe_tac, (chop_last (etac (reop rev xt6) 1)), Simp_tac 1,
    (dtac id_le_twice 1), (chop_last (etac (reop rev xt6) 1)), 
    (simp_tac (esimps add_ac) 1),
    (TRYALL (EVERY' [(dtac atomic_le_twice), (resolve_tac seq_atomic_singles),
      etac thin_rl, datac add_mono 1, (full_simp_tac (esimps add_ac)) ])) ]) ;

val _ = qed_goal_spec_mp "id_decompm_step1em" lctr.thy 
  "id_decomp (extrs rls) --> srl : rls --> srl : scrls --> \
    \ irl : rls --> irl : idrls --> pscmap (extend flr) irl = psc --> \
    \ inv_stepm (extrs rls) (extrs (multiples {srl})) psc"
  (fn _ => [ Safe_tac,
    (case_tac "seq_meet (snd irl) (snd srl) = 0" 1),
    (* case where the inversion rule is not in the {#A#} |- {#A#} part *)
    (case_tac "srl" 1), Clarify_tac 1,
    (rtac inv_stepm_disj_mul' 1), Full_simp_tac 1,
    (rtac refl 1), (rtac (singletonI RS extrs.I) 1),
    gen_cong_tac 1, (rtac refl 1), (rtac (thm "pair_collapse") 1),
    (rtac extrs_mono 1), Force_tac 1, nosg_tac 2,
    (* case where the inversion rule includes the {#A#} |- {#A#} part *)
    (etac swap 1), (rtac (mk_seq meet_commute RS trans) 1), (etac swap 1),
    (etac scrls.elims 1), (etac idrls.elims 1),
    (dtac (order_eq_refl RS seq_atom_disj_or_le) 1),
    Clarsimp_tac 1, Safe_tac, contr_tac 2,  (* level 23 *)
    (etac (le_add_diff_ex RS exE) 1), (rewtac id_decomp_def), 
    (eres_inst_tac [("x", "map (extend x) ps")] allE2 1), (etac impE 1),
    (clarsimp_tac (cesimps [inv_stepm.simps, invs_of_def]) 2), 
    (etac extrs.elims 2), (etac multiples.elims 2), Clarsimp_tac 2, 
    (case_tac "n" 2), Clarsimp_tac 2, 
    (rtac drs.derI 2), (chop_last (etac extrs.I 2)), 
    Force_tac 2, Simp_tac 2,
    Clarsimp_tac 2, (dtac subsetD 2), Fast_tac 2, (* level 40 *)
    (dtac (xt3 RS id_times_lem) 2), (rtac le_extend 2), 
    (dtac sym 2), (asm_full_simp_tac (esimps [extend_def]) 2),
    (etac (wk_adm_extrs RS wk_admD_alt) 2),
    (asm_full_simp_tac spsimps 2), (etac (le_plus1 RS xt6) 2), 
    (etac extrs.I 1), (simp_tac (esimps [extend_def]) 1), 
    (rtac conjI 1), atac 2, Fast_tac 1 ]) ;

val id_decompm_step1em' = idrls.I RS 
  reop (fn [a,b,c,d,e,f] => [e,a,f,d,c,b]) id_decompm_step1em ;

val _ = bind_thm ("id_decomp_step1em", id_decompm_step1em RS inv_step_m) ;

(*
val id_decomp_step1em' = idrls.I RS 
  reop (fn [a,b,c,d,e,f] => [e,a,f,d,c,b]) id_decomp_step1em ;
  *)
val id_decomp_step1em' = id_decompm_step1em' RS inv_step_m ;
val id_decomp_step1em'' = (id_decompm_step1em RS inv_step_mono RS inv_step_m) ; 

(* allow extra rules, since derivablility of premises not an issue,
  since no premises *)
val _ = qed_goal_spec_mp "no_prems_ext_step" lctr.thy
  "inv_step (derrec rls {}) irls (ps, c) --> ps = [] --> rls <= rls' --> \
    \ inv_step (derrec rls' {}) irls (ps, c)"
  (fn _ => [ (clarsimp_tac (cesimps [inv_step.simps]) 1),
    (datac subsetD 1 1), (etac der_mono 1), Auto_tac ]) ;

val _ = qed_goal_spec_mp "inv_stepm_empty" lctr.thy 
  "psc = (ps, c) --> irl = ([], ic) --> \
    \ inv_stepm rls (extrs {irl}) psc"
  (fn _ => [ (simp_tac (esimps [inv_stepm.simps, invs_of_def]) 1),
    Safe_tac, (etac extrs_E 1), (Clarsimp_tac 1) ]) ;

val _ = qed_goal_spec_mp "multiples_noprems" lctr.thy
  "psc : multiples {([], s)} --> (EX c. psc = ([], c))"
  (fn _ => [ (rtac impI 1), (etac multiples.elims 1), Force_tac 1 ]) ;
val multiples_nopremsE = multiples_noprems RS exE ;
    
val _ = qed_goal_spec_mp "inv_stepm_empty_mul" lctr.thy 
  "psc = (ps, c) --> irl = ([], ic) --> \
    \ inv_stepm rls (extrs (multiples {irl})) psc"
  (fn _ => [ (simp_tac (esimps [inv_stepm.simps, invs_of_def]) 1),
    Safe_tac, (etac extrs_E 1), (etac multiples_nopremsE 1), Force_tac 1 ]) ;

val inv_step_empty = refl RS inv_stepm_empty RS inv_step_m ; 
val inv_step_empty_mul = refl RS inv_stepm_empty_mul RS inv_step_m ;

(* can always invert identity *)
val _ = qed_goal_spec_mp "inv_stepm_extrs_id" lctr.thy
  "inv_stepm drls (extrs rls) psc --> \
    \ inv_stepm drls (extrs (rls Un idrls)) psc"
  (fn _ => [ (clarsimp_tac (cesimps [extrs_Un]) 1), (etac inv_stepm_Un 1), 
    (rtac (extrs_UN_single RS inv_stepm_UN') 1),
    (clarify_tac (ces [idrls.elims]) 1),
    (rtac ([surjective_pairing, refl] MRS inv_stepm_empty) 1) ]) ;

val _ = qed_goalw_spec_mp "id_decomp_step2" lctr.thy [id_decomp_def]
  "id_decomp rls --> wk_adm rls --> irl = (ips, {#A#} |- {#A#}) --> \
    \ irl : rls --> inv_step (derrec rls {}) (extrs {irl}) (ps, c)"
  (fn _ => [ (simp_tac (esimps [inv_step.simps, invs_of_def]) 1),
    Safe_tac, (etac allE2 1), mp_tac 1, (etac extrs_E 1),
    (rewtac wk_adm_def), (clarsimp_tac (cesimps [extend_def]) 1),
    (datac subsetD 1 1), Auto_tac ]) ;

val _ = qed_goal_spec_mp "scrls_same_or_disj" lctr.thy
  "(psa, ca) : scrls --> (psb, cb) : scrls --> ca = cb | seq_meet ca cb = 0"
  (fn _ => [ (REPEAT (rtac impI 1)),
    (REPEAT (etac scrls.elims 1)),
    (REPEAT (dtac atom_a_or_s 1)), Safe_tac,
    (ALLGOALS (full_simp_tac (esimps [seq_meet_def, seq_zero_def,
      ms_meet_singletons] addsplits [split_if_asm]))),
    (TRYALL (force_tac (cesimps [ms_meet_pg]))) ]) ; 

val id_decomp_step2' = [asm_rl, wk_adm_extrs, refl] MRS id_decomp_step2 ;

val _ = qed_goal_spec_mp "inv_stepm_scrls" lctr.thy 
  "extrs {srl} = rls --> rl : rls --> \
    \ srl : scrls --> irl : scrls --> (snd srl = snd irl --> srl = irl) --> \
    \ inv_stepm rls (extrs {irl}) rl"
  (* irl = inversion rule, srl = rule in last step of proof *)
  (fn _ => [ (Clarify_tac 1), (case_tac "irl" 1), (case_tac "srl" 1),
    (Clarsimp_tac 1), (datac scrls_same_or_disj 1 1),
    (etac disjE 1), mp_tac 1,
    (* case where inversion rule and last step rule are the same *)
    (rtac inv_stepm_same 1), Fast_tac 1,
    (* case where different rules from scrls are used *)
    (etac inv_stepm_disj 1), (TRYALL (ares_tac [refl])) ]) ;

val _ = qed_goal_spec_mp "inv_stepm_scrls_mul" lctr.thy 
  "extrs {srl} = rls --> rl : rls --> srl : scrls --> irl : scrls --> \
      \ (snd srl = snd irl --> srl = irl) --> \
    \ inv_stepm rls (extrs (multiples {irl})) rl"
  (* irl = inversion rule, srl = rule in last step of proof *)
  (fn _ => [ (Clarify_tac 1), (case_tac "irl" 1), (case_tac "srl" 1),
    (Clarsimp_tac 1), (datac scrls_same_or_disj 1 1),
    (etac disjE 1), mp_tac 1,
    (* case where inversion rule and last step rule are the same *)
    (rtac inv_stepm_same_mul 1), Fast_tac 1, Fast_tac 1,
    (* case where different rules from scrls are used *)
    (etac inv_stepm_disj_mul 1), (TRYALL (ares_tac [refl])) ]) ;

val _ = bind_thm ("inv_step_scrls", reop (fn [a,b,c,d,e,f] => [a,b,f,c,d,e])
  (inv_stepm_scrls RS inv_step_mono RS inv_step_m)) ;
val inv_stepm_scrls_mul' = reop (fn [a,b,c,d,e,f] => [a,b,f,c,d,e])
  (inv_stepm_scrls_mul RS inv_step_mono) ;
val _ = bind_thm ("inv_step_scrls_mul", inv_stepm_scrls_mul' RS inv_step_m) ;

val _ = qed_goal_spec_mp "inv_stepm_ctrrls" lctr.thy 
  "crl : extrs (UNION UNIV ctrrls) --> irl : scrls --> \
    \ crls = extrs (UNION UNIV ctrrls) --> \
    \ inv_stepm crls (extrs (multiples {irl})) crl"
  (fn _ => [ (clarify_tac (ces [extrs.elims]) 1),
    (* case where the contraction rule and the rule to be inverted 
      have different conclusions *)
    (case_tac "snd irl = ba" 1),
    (rtac inv_stepm_scrls_mul' 2), atac 6, 
    (etac (ctr_scrls RS subsetD) 5), Force_tac 5,
    (rtac refl 2), (etac (singletonI RS extrs.I) 2),
    (rtac extrs_mono 2), Fast_tac 2, 
    (* case where the contraction rule and the rule to be inverted 
      have same conclusions *)
    (clarsimp_tac (ces [extrs.elims, multiples.elims],
      esimps [inv_stepm.simps, invs_of_def]) 1),
    (* case of times 0 irl : multiples {irl} *)
    (case_tac "n" 1), 
    (clarsimp_tac (cesimps [extend_def]) 1), 
    (rtac drs.derI 1), 
    (etac (UNIV_I RS UN_I RS extrs.I) 1),
    (force_tac (cesimps [extend_def]) 1), 
    (clarsimp_tac (cesimps [drs.all]) 1), 
    (rtac drs.dpI 1), (rtac UnI1 1), (etac imageI 1),
    (dtac ctr_double 1), (clarsimp_tac (claset (), spsimps) 1), 
    (rtac (dd1ths RS der_mono) 1), rtac subset_refl 2, Clarify_tac 2,
    (rtac exI 2), rtac conjI 2,
    (res_inst_tac [("flr", "flra")] extrs.I 3), 
    (res_inst_tac [("n", "Suc (Suc nat)")] multiples.I 3), 
    rtac singletonI 3, (simp_tac spsimps 3),
    (thin_tac "Not ?P" 2), Force_tac 2,
    (simp_tac (esimps [extend_def, extrs_UN RS sym, ctrrls_derl_iff]) 1),
    (rtac (seq_singles_le RS xt6) 1),
    (rtac (seq_singles_double RS sym RS xt3) 1),
    (rtac seq_singles_mono 1), (simp_tac spsimps 1) ]) ;

val _ = bind_thm ("inv_step_ctrrls", 
  refl RSN (3, inv_stepm_ctrrls) RS inv_step_mono RS inv_step_m) ;

val wk_prem = [subset_refl, drs.dpI] MRS wk_admD_alt_ps  ;

(* weakening rule allows inv_step *)
val _ = qed_goal_spec_mp "inv_stepm_wk" lctr.thy 
  "irls <= scrls --> (ps, c) : extrs wkrls --> \
    \ inv_stepm (extrs wkrls) (extrs irls) (ps, c)"
  (fn _ => [ (clarsimp_tac (ces [extrs.elims],
      esimps [inv_stepm.simps, invs_of_def]) 1),
    (datac subsetD 1 1), (etac scrls.elims 1),
    (dtac atom_a_or_s 1), (case_tac "flr" 1), (case_tac "flra" 1), 
    (safe_tac (ces [wkrls.elims])),
    (ALLGOALS (clarsimp_tac (cesimps [extend_def]))),
    (TRYALL (EVERY' [(dtac (add_eq_conv_ex' RS iffD1)),
      (etac disjE THEN_ALL_NEW Clarify_tac),
      (rtac (insertI1 RS wk_prem)), MSSG Simp_tac])), 
    (TRYALL (EVERY' [
      (rtac (insertI2 RS wk_prem)), Simp_tac, rtac exI, rtac rev_conjI, 
      (etac extrsI2), (simp_tac (esimps [extend_def])), Fast_tac,
      Clarsimp_tac, etac imageI, MSSG (simp_tac (esimps [extend_def])) ])) ]) ;

val amth = prove_goal lctr.thy "t + x + (a |- b) >= t + (a |- b)" 
  (fn _ => [ (asm_simp_tac (esimps [] addsimprocs pmgsps) 1) ]) ;

val dsn_ss = esimps [extend_def] delsimps [times.S] ;

val tstacs = [
  (rtac (insertI2 RS wk_prem)), rtac CollectI, rtac exI,
  rtac rev_conjI, (etac (multiples.I RS extrsI2)), 
  (simp_tac dsn_ss), rtac conjI, rtac refl,
  (force_tac (claset (), dsn_ss addsimps [seq_times])),
  (force_tac (claset (), dsn_ss)), (MSSG (simp_tac dsn_ss)) ] ;

val _ = qed_goal_spec_mp "inv_stepm_wkm" lctr.thy 
  "irls <= scrls --> (ps, c) : extrs wkrls --> \
    \ inv_stepm (extrs wkrls) (extrs (multiples irls)) (ps, c)"
  (fn _ => [ (clarsimp_tac (ces [extrs.elims, multiples.elims],
      esimps [inv_stepm.simps, invs_of_def]) 1),
    (case_tac "n" 1), Clarsimp_tac 1, 
    (rtac drs.derI 1), (etac extrs.I 1),
    (rtac pscmap_def 1), (simp_tac (esimps [drs.all]) 1), 
    rtac subsetI 1, rtac drs.dpI 1, etac UnI1 1, nosg_tac 2,

    (datac subsetD 1 1), (etac scrls.elims 1),
    (dtac atom_a_or_s 1), (case_tac "flr" 1), (case_tac "flra" 1), 
    (safe_tac (ces [wkrls.elims])),
    (ALLGOALS (clarsimp_tac (cesimps [extend_def, seq_times]))),
    (* for sgs 1 & 4, to cope with case 
      times nat {#x#} + {#x#} + multiset2a = {#A#} + multiset2 *)
    (TRYALL (EVERY' [ (dtac times_single_add1'),
      etac disjE THEN_ALL_NEW Clarify_tac, EVERY' tstacs ])),
    (fold_goals_tac [mk_eq times.S]), (TRYALL (EVERY' tstacs)) ]) ;

val _ = bind_thm ("inv_step_wk", inv_stepm_wk RS inv_step_mono RS inv_step_m) ;
val _ = bind_thm ("inv_step_wkm", inv_stepm_wkm RS inv_step_mono RS inv_step_m);

val inv_step_wkm' = reop (fn [a,b,c,d] => [b,a,c,d]) 
  (sym RSN (2, extrs.I) RSN (2, inv_step_wkm)) ;
val inv_step_ctrrls' = [UNIV_I RS UN_I, sym] MRS extrs.I RS inv_step_ctrrls ;

val rsD' = prove_goal HOL_Gen.thy
  "x : A --> A <= B --> x : A & x : B" fn_at RS mp RS mp ;

(* gen_inv : generalise inversion property : 
  if rule set rls is in iscrls (single conclusion rules & axiom),
    rls has the unique-conclusion property,
    extrs rls has the identity decomposition property,
    the all rules in extrs rls are invertible *)

(* first try only inverting single_conclusion rules *)

val wklem = reop rev
  ([asm_rl, extrs.I, Un_upper2 RS extrs_mono] MRS inv_step_wk) ;
val wklemc = reop rev ([asm_rl, extrs.I, 
  [Un_upper2, Un_upper1] MRS subset_trans RS extrs_mono] MRS inv_step_wk) ;
val wklemm = reop rev ([asm_rl, extrs.I, 
  [Un_upper2, Un_upper1] MRS subset_trans RS extrs_mono] MRS inv_step_wkm) ;

(* separate theorem from gen_inv'' as must have both wk rules or neither,
  for inversion of not-intro rules *)
val () = qed_goal_spec_mp "gen_inv_wk3" lctr.thy 
  "rlsnw <= iscrls --> rls = rlsnw Un wkrls --> unique_concl rlsnw --> \
    \ id_decomp (extrs rlsnw) --> r : rlsnw --> \
    \ Ball (extrs rls) (inv_step (derrec (extrs rls) {}) (extrs {r}))"
  (fn _ => [ (safe_tac (ces [extrs_E])), 
    (REPEAT_FIRST (datac rsD' 1)),
    (safe_tac (ces [iscrls.elims, idrls.elims])), 
    (* case of last rule weakening *)
    (TRYALL (eatac wklem 1 THEN' Fast_tac)),
    (* invert the identity axiom *)
    (TRYALL (rtac (refl RS inv_step_empty))), nosg_tac 3,
    (* inversion step where last rule uses identity axiom, *)
    (TRYALL (EVERY' [(datac id_decomp_step1e' 4),
      (etac no_prems_ext_step), Force_tac, (rtac extrs_mono), Fast_tac])), 
    (* last rule used and inversion rule are rules from scrls *)
    (rtac inv_step_scrls 1),
    (TRYALL (ares_tac [refl]) THEN etac (singletonI RS extrs.I) 1), 
    (rtac extrs_mono 1), Force_tac 1, 
    (clarsimp_tac (cesimps [unique_concl_def]) 1), 
    (etac allE2 1), (etac impE 1), (rtac exI 1), 
    (eatac conjI 1 1), Force_tac 1 ]) ;

val () = qed_goal_spec_mp "gen_inv_wk''" lctr.thy 
  "rlsnw <= iscrls --> rls = rlsnw Un wkrls --> unique_concl rlsnw --> \
    \ id_decomp (extrs rlsnw) --> r : rlsnw --> (ps, c) : extrs {r} --> \
    \ c : derrec (extrs rls) {} --> ps : dersrec (extrs rls) {}"
  (fn _ => [ (REPEAT (rtac impI 1)), (rtac gen_inv_step 1), TRYALL atac,
    Clarify_tac 1, (etac gen_inv_wk3 1), rtac refl 1, TRYALL atac ]) ;

(* now try adding in contraction rules - again, must be all or none,
  require all multiples, all extrs, of some rule to invert *)
val () = qed_goal_spec_mp "gen_inv_wc3" lctr.thy 
  "rlsnw <= iscrls --> rls = rlsnw Un wkrls Un UNION UNIV ctrrls --> \
    \ unique_concl rlsnw --> id_decomp (extrs rlsnw) --> r : rlsnw --> \
    \ Ball (extrs rls) (inv_step (derrec (extrs rls) {}) \
      \ (extrs (multiples {r})))"
  (fn _ => [ (safe_tac (ces [extrs_E])), 
    (REPEAT_FIRST (datac rsD' 1)),
    (safe_tac (ces [iscrls.elims, idrls.elims])), 
    (* case of last rule weakening *)
    (* doesn't work as requires multiples {(ac, bc)} <= scrls,
      now have inv_step_wkm *)
    (TRYALL (eatac wklemm 1 THEN' Fast_tac)),
    (* invert the identity axiom *)
    (TRYALL (rtac (refl RS inv_step_empty_mul))), nosg_tac 4,
    (* inversion step where last rule uses identity axiom, *)
    (EVERY' [(datac id_decomp_step1em' 4),
      (etac no_prems_ext_step), Force_tac, (rtac extrs_mono), Fast_tac] 1), 

    (* last rule used and inversion rule are rules from scrls *)
    (rtac inv_step_scrls_mul 1),
    (TRYALL (ares_tac [refl]) THEN etac (singletonI RS extrs.I) 1), 
    (rtac extrs_mono 1), Force_tac 1, 
    (clarsimp_tac (cesimps [unique_concl_def]) 1), 
    (etac allE2 1), (etac impE 1), (rtac exI 1), 
    (eatac conjI 1 1), Force_tac 1, (* level 17 *)
    
    (* now, when the last rule used is contraction *)
    (rtac inv_step_ctrrls 1), 
    (rtac extrs.I 1), atac 2, Fast_tac 1, atac 1, 
    (rtac extrs_mono 1), Fast_tac 1 ]) ;

val () = qed_goal_spec_mp "gen_inv_wc''" lctr.thy 
  "rlsnw <= iscrls --> rls = rlsnw Un wkrls Un UNION UNIV ctrrls --> \
    \ unique_concl rlsnw --> id_decomp (extrs rlsnw) --> \
    \ r : rlsnw --> (ps, c) : extrs (multiples {r}) --> \
    \ c : derrec (extrs rls) {} --> ps : dersrec (extrs rls) {}"
  (fn _ => [ (REPEAT (rtac impI 1)), (rtac gen_inv_step 1), TRYALL atac,
    Clarify_tac 1, (etac gen_inv_wc3 1), rtac refl 1, TRYALL atac ]) ;

val () = qed_goal_spec_mp "gen_inv3" lctr.thy 
  "rls <= iscrls --> unique_concl rls --> id_decomp (extrs rls) --> \
    \ r : rls --> \ 
    \ Ball (extrs rls) (inv_step (derrec (extrs rls) {}) (extrs {r}))"
  (fn _ => [ 
    (clarify_tac (ces [extrs_E]) 1), 
    (REPEAT (datac rsD' 1 1)),
    (safe_tac (ces [iscrls.elims, idrls.elims])), 
    (* invert the identity axiom - trivial, why do we need this? *)
    (TRYALL (rtac (refl RS inv_step_empty))),
    (* inversion step where last rule uses identity axiom *)
    (eatac id_decomp_step1e' 4 1),
    (* last rule used and inversion rule are rules from scrls *)
    (rtac inv_step_scrls 1),
    (TRYALL (ares_tac [refl]) THEN etac (singletonI RS extrs.I) 1), 
    (rtac extrs_mono 1), Force_tac 1, 
    (clarsimp_tac (cesimps [unique_concl_def]) 1), 
    (etac allE2 1), (etac impE 1), (rtac exI 1), 
    (eatac conjI 1 1), Force_tac 1 ]) ;

val () = qed_goal_spec_mp "gen_inv''" lctr.thy 
  "rls <= iscrls --> unique_concl rls --> id_decomp (extrs rls) --> \
    \ r : rls --> (ps, c) : extrs {r} --> \
    \ c : derrec (extrs rls) {} --> ps : dersrec (extrs rls) {}"
  (fn _ => [ (REPEAT (rtac impI 1)), (rtac gen_inv_step 1), TRYALL atac,
    Clarify_tac 1, (eatac gen_inv3 4 1) ]) ;

val gen_inv' = reop (rev o (fn a :: b :: r => b :: a :: r) o rev) gen_inv'' ;

val () = qed_goal_spec_mp "gen_inv" lctr.thy 
  "rls <= iscrls --> unique_concl rls --> id_decomp (extrs rls) --> \
    \ Ball (extrs rls) (inv_rl (extrs rls))"
  (fn _ => [ (clarsimp_tac (cesimps [inv_rl.simps]) 1),
    (dtac (extrs_UN_single RS equalityD2') 1), etac UN_E 1,
    (rtac gen_inv'' 1), TRYALL atac ]) ;
    
val () = qed_goal_spec_mp "gen_inv_wk" lctr.thy 
  "rls <= iscrls --> rlsw = rls Un wkrls --> unique_concl rls --> \
    \ id_decomp (extrs rls) --> Ball (extrs rls) (inv_rl (extrs rlsw))"
  (fn _ => [ (clarsimp_tac (cesimps [inv_rl.simps]) 1),
    (dtac (extrs_UN_single RS equalityD2') 1), etac UN_E 1,
    (rtac gen_inv_wk'' 1), TRYALL atac, rtac refl 1 ]) ;
    
val () = qed_goal_spec_mp "gen_inv_wc" lctr.thy 
  "rlsnw <= iscrls --> rls = rlsnw Un wkrls Un UNION UNIV ctrrls --> \
    \ unique_concl rlsnw --> id_decomp (extrs rlsnw) --> \
    \ Ball (extrs (multiples rlsnw)) (inv_rl (extrs rls))"
  (fn _ => [ (clarsimp_tac (cesimps [inv_rl.simps]) 1),
    (dtac (em_UN_single RS equalityD2') 1), etac UN_E 1,
    (rtac gen_inv_wc'' 1), TRYALL atac, rtac refl 1 ]) ;
    
(** similar proof using modified inversion property mod_inv **)
val _ = qed_goal_spec_mp "gen_mod_inv_step'" lctr.thy
  "Ball rls (mod_inv_step (derrec rls {}) irls f) --> c : derrec rls {} --> \
    \ mod_inv_prop (derrec rls {}) irls (seq_filter f) c"
  (fn _ => [ (REPEAT (rtac impI 1)),
    (etac drs.inductr 1), Fast_tac 1,
    (force_tac (cesimps [mod_inv_step.simps, drs.all]) 1) ]) ;
    
mod_inv_step.simps ;
mod_inv_prop_def ;
inv_step.simps ;
(* 
set show_sorts ;
reset show_sorts ;
*)
      
(* invertibility in the presence of an explicit contraction rule *)
(*
ctr_adm_def ;
ctr_admd_def ;
ctr_adm_seq_def ;
ctr_adm_prop_def ;
ctr_adm_step_def ;
ctr_adm_step1.simps ;
ctr_adm_step_simp ;
*)

(** gen_ctr_adm : admissibility of contraction, generalised version :
  for a well-founded subformula relation, if 
  rls are in (single-conclusion or axiom),
  rls have the subformula property, except where conclusion in premise,
  extrs rls are invertible, then extrs rls have contraction admissibility **)

val () = qed_goal_spec_mp "ctr_adm_alt" lctr.thy  
  "ctr_adm rls A = Ball (derrec rls {}) (ctr_adm_prop (derrec rls {}) A)"
  (fn _ => [ (force_tac (cesimps [ctr_adm_prop_def, ctr_adm_def]) 1) ]) ;

val () = qed_goal_spec_mp "ctr_admd_alt" lctr.thy 
  "ctr_admd derivs A = (ALL seq: derivs. ctr_adm_prop derivs A seq)"
  (fn _ => [ (force_tac (cesimps
      [ctr_admd_def, ctr_adm_seq_def, ctr_adm_prop_def]) 1) ]) ;

(* why does this not work ?? 
val () = qed_goal_spec_mp "ctr_adm_step_gen" lctr.thy 
  "ctr_adm_step sub derivs (ps, c) dt = \
    \ gen_step (ctr_adm_prop derivs) dt sub derivs (ps, c)"
  (fn _ => [ 
    (simp_tac (esimps [ctr_adm_step_def, gen_step.simps, ctr_admd_alt]) 1),
    (simp_tac (esimps [ctr_adm_step1.simps, ctr_adm_prop_def]) 1) ]) ;
*)

val () = qed_goal_spec_mp "ctr_adm_step_gen" lctr.thy 
  "ctr_adm_step sub derivs (ps, c) dt --> \
    \ gen_step (ctr_adm_prop derivs) dt sub derivs (ps, c)"
  (fn _ => [ 
    (simp_tac (esimps [ctr_adm_step_def, gen_step.simps, ctr_admd_alt]) 1),
    (simp_tac (esimps [ctr_adm_step1.simps, ctr_adm_prop_def]) 1) ]) ;

(* can we use this - prove that all the rules satisfy ctr_adm_step ? *)
val () = qed_goalw_spec_mp "gen_ctr_adm_step" lctr.thy [mk_eq ctr_adm_alt]
  "A : wfp sub --> (ALL dt. ALL (ps, concl) : rules. \
    \ ctr_adm_step sub (derrec rules {}) (ps, concl) dt) --> ctr_adm rules A"
  (fn _ => [ (Clarify_tac 1), 
    (dtac gen_step_lem 1), atac 2, atac 2, 
    (etac all_forward 1), (etac ball_match 1),
    Clarify_tac 1, (etac ctr_adm_step_gen 1) ]) ;

val xlem = reop rev (diff_le RS xt6 RS (le_add_diff_iff RS iffD2) RS xt3) ;

(* a simple case - contracting from sequent containing axiom *)
val () = qed_goal_spec_mp "ax_ctr_adm_step1" lctr.thy 
  "({#B#} |- {#B#}) <= concl --> extrs idrls <= rules --> \
    \ ctr_adm_step1 (derrec rules {}) (ps, concl) As"
  (fn _ => [ (clarsimp_tac (cesimps [ctr_adm_step1.simps]) 1),
    (thin_tac "Ball ?S ?P" 1), 
    (case_tac "concl" 1), (case_tac "As" 1), Clarsimp_tac 1,
    (REPEAT1 (dtac sum_le 1)), (REPEAT1 (datac xlem 1 1)), 
    (safe_tac (ces [atomic_single RS atomic_le_sumD RS disjE])),
    (REPEAT_FIRST (datac xt6 1)),
    (REPEAT_FIRST (dtac (single_le RS iffD1))),
    (TRYALL (eatac extrs_id_ax' 2)) ]) ;
    
val () = qed_goalw_spec_mp "ax_ctr_adm_step" lctr.thy [ctr_adm_step_def]
  "({#B#} |- {#B#}) <= concl --> extrs idrls <= rules --> \
    \ ctr_adm_step sub (derrec rules {}) (ps, concl) A"
  (fn _ => [ Clarify_tac 1, (eatac ax_ctr_adm_step1 1 1) ]) ; 

(* need extrs ({#A#} |- {#A#}) to be in rules *)
val () = qed_goal_spec_mp "id_ctr_adm_step" lctr.thy 
  "rl : idrls --> pscmap (extend flr) rl = erl --> \
    \ extrs {rl} <= rules --> ctr_adm_step sub (derrec rules {}) erl dt"
  (fn _ => [ (clarsimp_tac (ces [idrls.elims],
      esimps [extend_def, ctr_adm_step_simp]) 1),
    (dtac ax_step2 1), (dtac le_pm_assoc 1), 
    (rtac mem_same 1), atac 2, 
    (rtac drs.derINil 1), (etac subsetD 1),
    (rtac extrs.I 1), Fast_tac 1,
    (simp_tac (esimps [extend_def]) 1) ]) ;

(* now to do a similar thing for a rule in scrls *)

val the = zero_var_indexes ([singletonI, refl] MRS extrs.I RS rev_subsetD) ;

val _ = qed_goal_spec_mp "ctxt_ctr_adm_step1" lctr.thy 
  "pg_meet concl As = 0 --> rl = (ps, concl) --> \
    \ pscmap (extend flr) rl = erl --> extrs {rl} <= rules --> \
    \ ctr_adm_step1 (derrec rules {}) erl As"
  (fn _ => [ (clarsimp_tac (cds [le_add_diff_ex],
	esimps [extend_def, ctr_adm_step1.simps]) 1),
      (ftac (le_plus1 RSN (2, (nth (xtrans, 2)))) 1),
      (rewtac (mk_eq add_assoc)),
      (REPEAT1 (datac (rotate_prems 1 pg_disj_le_add) 1 1)),
      (dtac le_add_diff_ex 1), Clarify_tac 1,
      (subgoal_tac "flr = As + As + Da" 1),
      (MSSG (full_simp_tac spsimps) 2),
      (simp_tac spsimps 1), (rtac drs.derI 1), 
      (forw_inst_tac [("flr", "As + Da")] the 1), 
      hyp_subst_tac 1, (MSSG (full_simp_tac spsimps) 1),
      (full_simp_tac (spsimps addsimps [drs.all, image_subset_iff]) 1),
      (etac ball_match 1), Clarify_tac 1, 
      (force_tac (claset (), spsimps) 1) ]) ;

val cs1 = tacthm (TRYALL (rtac refl)) ctxt_ctr_adm_step1 ;
val cs2 = simplify (esimps [ctr_adm_step1.simps]) cs1 ;
val cs3 = meta_std cs2 ;

val _ = qed_goalw_spec_mp "ctr_adm_step_lem" lctr.thy [ctr_adm_step_def]
  "(ALL As. pg_meet concl As = 0 --> ctr_adm_step1 derivs erl As) --> \
    \ ~ (A :# ms_of_seq concl) --> ctr_adm_step sub derivs erl A"
  (fn _ => [ Clarify_tac 1, (etac allE 1), (etac mp 1),
    (case_tac "As" 1), (case_tac "concl" 1),
    (clarsimp_tac (cesimps 
      [seq_meet_def, symmetric seq_meet_pg, seq_zero_def]) 1),
    (dtac ([atomic_single, sym] MRS atomic_eq_sum) 1), Safe_tac,
    (ALLGOALS (clarsimp_tac 
      (cesimps (ms_meet_singleton' :: ms_meet_join_0)))) ]) ;

val _ = qed_goal_spec_mp "ctxt_ctr_adm_step" lctr.thy 
  "~ (A :# ms_of_seq concl) --> rl = (ps, concl) --> \
    \ pscmap (extend flr) rl = erl --> extrs {rl} <= rules --> \
    \ ctr_adm_step sub (derrec rules {}) erl A"
  (fn _ => [ (Clarify_tac 1), (rtac ctr_adm_step_lem 1), Clarify_tac 1,
    (rtac ctxt_ctr_adm_step1 1), (TRYALL (ares_tac [refl])), Simp_tac 1 ]) ;

val meet_iff_le' = fcomm_all pg_meet_commute (transfer pmgl.thy meet_iff_le) ;

(* now similarly for extcs *)
(* can do better for extcs *)
val _ = qed_goal_spec_mp "ctxt_ctr_adm_step1_cs" lctr.thy 
  "pg_meet concl (As + As) <= As --> rl = (ps, concl) --> \
    \ (ps, extend flr concl) = erl --> extcs {rl} <= rules --> \
    \ ctr_adm_step1 (derrec rules {}) erl As"
  (fn _ => [ (clarsimp_tac (cesimps [extend_def, ctr_adm_step1.simps]) 1),
    (rtac drs.derI'' 1), (etac subsetD 1),
    (res_inst_tac [("flr", "flr - As")] (singletonI RS extcs.I) 1), 
    (simp_tac (esimps [extend_def]) 1),
    (rtac (refl RS conjI) 1), Fast_tac 2, (rtac le_pm_assoc 1),
    (cut_inst_tac [("A", "concl"), ("B", "flr"), ("C", "As + As")]
      meet_plus_ge 1) ,
    (chop_tac 1 (dtac (meet_iff_le RS iffD2) 1)), Clarsimp_tac 1,
    (chop_tac 1 (datac le_sub_le 1 1)), Clarsimp_tac 1, 
    (etac (reop rev xt6) 1), Simp_tac 1 ]) ;

val _ = qed_goalw_spec_mp "ctxt_ctr_adm_step_cs" lctr.thy [ctr_adm_step_def]
  "~ ({#A#} + {#A#} <= ms_of_seq concl) --> rl = (ps, concl) --> \
    \ (ps, extend flr concl) = erl --> extcs {rl} <= rules --> \
    \ ctr_adm_step sub (derrec rules {}) erl A"
  (fn _ => [ (clarsimp_tac (cesimps [c2lem]) 1),
    (rtac ctxt_ctr_adm_step1_cs 1), (TRYALL (ares_tac [refl])),
    (case_tac "concl" 1), (dtac ms_of_single 1), Safe_tac, 
    (ALLGOALS (clarsimp_tac (cesimps 
      ([ms_le_def, symmetric seq_meet_pg, seq_meet_def] @ ms_meet_join_0)))),
    Safe_tac, 
    (ALLGOALS (clarsimp_tac (cesimps [count_ms_meet, meet_count_def]))),
    ALLGOALS arith_tac ]) ;

val () = qed_goal_spec_mp "sc_ctr_adm_step2f" lctr.thy 
  "rl = (ps, concl) --> rl : scrls --> pscmap (extend flr) rl = erl --> \
    \ forget (subfml_cp_prop sub rl & \
      \ forget (extrs {rl} <= Collect (inv_rl rules)) | \
    \ (ALL p : set ps. concl <= p)) --> \
    \ extrs {rl} <= rules --> ctr_adm_step sub (derrec rules {}) erl dt"
  (fn _ => [ (clarsimp_tac (cesimps [ctr_adm_step_simp]) 1),
    (etac scrls.elims 1), (dtac le_add_diff_ex 1), 
    (etac (tacthm (dtac sym 2) exE) 1), (Clarsimp_tac 1),
    (subgoal_tac "concl <= As | concl <= D & ~ (concl <= As)" 1), 
    (asm_full_simp_tac (esimps [extend_def]) 2),
    (dtac le_plus_eq1 2),
    (fast_tac (cds [reop rev (order_eq_refl RS atom_le_union')]) 2),
    
    (ftac subsetD 1), (rtac (singletonI RS in_extrs) 1), 
    (etac disjE 1), (thin_tac "All ?P" 2), (etac thin_rl 2), (* level 14 *)

    (* case - principal formula of rule in context of contraction *)
    (dtac ([sym, seq_eq_refl] MRS trans) 2), Asm_simp_tac 2, 
    (rtac cs3 2), atac 3, atac 3,
    (etac ([sym, le_plus1] MRS nth (xtrans, 2)) 3),
    (rtac (seq_atomic_def RS iffD2 RS atomicD RS make_pos_rule') 2),
    (Force_tac 2), (Force_tac 2), 

    (* case - principal formula of rule is principal formula of contraction *)
    (subgoal_tac "concl = As" 1), (etac lssl 2),
    (asm_full_simp_tac (esimps [seq_size_ms]) 2),
    (subgoal_tac "flr = As + D" 1), 
    (ALLGOALS (clarsimp_tac (cesimps (extend_def :: plus_ac)))),
    (rtac drs.derI' 1),
    (forw_inst_tac [("flr", "D")] the 1), 
    (MSSG (full_simp_tac spsimps) 1), (* level 34 *)

    (clarsimp_tac (cesimps [extend_def, drs.all, image_subset_iff]) 1),
    (dtac forgetD 1), (etac disjE 1),
    
    (* case - principal formula in all premises *)
    (REPEAT1 (datac bspec 1 2)), (etac conjE 2), (etac impE 2),
    etac order_trans 2, (rtac le_plus2 2), 
    (MSSG (full_simp_tac (esimps plus_ac)) 2), (* level 44 *)

    (etac conjE 1), 
    (rewtac (mk_eq subfml_cp_prop.simps)), 
    (REPEAT1 (datac bspec 1 1)), (etac disjE 1),
    
    (* case - principal formula in premise *)
    (etac conjE 1), (etac impE 1),
    etac order_trans 1, (rtac le_plus2 1), 
    (MSSG (full_simp_tac (esimps plus_ac)) 1), (* level 44 *)

    (* case - formulae in premise are subformulae of principal formula *)
    (dtac conjunct1 1), 
    (* use invertibility *)
    (subgoal_tac "(x + x) + D : derrec rules {}" 1),
    (full_simp_tac (esimps [forget_def]) 2),
    (chop_tac 1 (dres_inst_tac [("flr", "D + x")] the 2)), 
    (full_simp_tac (esimps [extend_def, inv_rl.simps]) 2),
    (etac impE 2), (MSSG (full_simp_tac spsimps) 2), 
    (full_simp_tac (spsimps addsimps [drs.all, image_subset_iff]) 2), 
    Fast_tac 2, (* level 54 *)

    (rtac multi_ctr 1), (atac 2), (Clarify_tac 1),
    (etac allE 1), (rewtac ctr_adm_def'), (etac mp 1),
    (force_tac (claset (), esimps [ms_mem_def] addsplits [split_if_asm]) 1)]);

val sc_ctr_adm_step2 = rewrite_rule [forget_def] sc_ctr_adm_step2f ;

val _ = bind_thm ("sc_ctr_adm_step", 
  tacsthm [rtac disjI1 4, rtac conjI 4] sc_ctr_adm_step2) ;

val () = qed_goal_spec_mp "gen_ctr_adm" lctr.thy 
  "A : wfp sub --> rls <= iscrls --> Ball rls (subfml_cp_prop sub) --> \
    \ Ball (extrs rls) (inv_rl (extrs rls)) --> \
    \ ctr_adm (extrs rls) A" 
  (fn _ => [ (safe_tac (ces [gen_ctr_adm_step, extrs.elims])),
    (datac subsetD 1 1), (etac iscrls.elims 1),
    (etac id_ctr_adm_step 1), Fast_tac 1,
    (Clarify_tac 2),
    (rtac (refl RS sc_ctr_adm_step) 2), atac 2, atac 2,
    (safe_tac (ces [extrs_mono', bspec])) ]) ;

(* contraction admissibility for system including weakening rules *)
val () = qed_goal_spec_mp "wk_ctr_adm_step" lctr.thy 
  "rl = (ps, concl) --> rl : wkrls --> pscmap (extend flr) rl = erl --> \
    \ wk_adm rules --> ctr_adm_step sub (derrec rules {}) erl dt"
  (fn _ => [ (clarsimp_tac (cesimps [ctr_adm_step_simp]) 1),
    (thin_tac "All ?P" 1),
    (etac wk_altE 1), Clarsimp_tac 1, 
    (case_tac "As = concl" 1),
    (* case - contracting the formula just introduced by weakening *)
    (force_tac (cesimps [extend_def]) 1),
    (* case - weakening in the context of the contraction *)
    (clarsimp_tac (cesimps [extend_def]) 1),
    (etac impE 1),
    (etac (pm_le_eq RS iffD2 RS xt4) 1),
    (rtac ([le0_eq, pg_disj_minus] MRS trans RS iffD1) 1),
    (rtac (meet_plus_ge RSN (2, xt3)) 1),
    (rtac (conjI RS (plus_eq0 RS iffD2 RS sym)) 1),
    (TRYALL (rtac (tacthm (atac 4) (atoms_disj_or_eq RS disjE)))),
    (TRYALL (clarsimp_tac (cesimps [seq_atomic_def, seq_size_ms]))),
    (etac wk_admD_alt 1), (chop_last (atac 1)),
    (rtac le_sub_le 1), Auto_tac ]) ;

val () = qed_goal_spec_mp "gen_ctr_adm_step_inv" lctr.thy 
  "epsc : extrs lrls --> lrls <= iscrls --> \
    \ extrs lrls <= drls --> Ball lrls (subfml_cp_prop sub) --> \
    \ Ball (extrs lrls) (inv_rl drls) --> \
    \ ctr_adm_step sub (derrec drls {}) epsc A"
  (fn _ => [ (safe_tac (ces [extrs.elims])),
    (datac subsetD 1 1), (etac iscrls.elims 1),
    (etac id_ctr_adm_step 1), Fast_tac 1,
    (etac xt6 1), (rtac extrs_mono 1), Fast_tac 1,
    (Clarify_tac 1),
    (rtac (refl RS sc_ctr_adm_step) 1), atac 1, atac 1,
    (safe_tac (ces [subsetD, extrs_mono', bspec])) ]) ;

val () = qed_goal_spec_mp "gen_ctr_adm_wk" lctr.thy 
  "A : wfp sub --> rls <= iscrls --> rlsw = rls Un wkrls --> \
    \ Ball rls (subfml_cp_prop sub) --> \
    \ Ball (extrs rls) (inv_rl (extrs rlsw)) --> ctr_adm (extrs rlsw) A" 
  (fn _ => [ 
    (safe_tac (ces [gen_ctr_adm_step, extrs_Un RS equalityD1' RS UnE])),
    (rtac gen_ctr_adm_step_inv 1), TRYALL atac, 
    (rtac (Un_upper1 RS extrs_mono) 1),
    (safe_tac (ces [extrs.elims])), 
    (rtac wk_ctr_adm_step 1), 
    (TRYALL (UNIQUE o atac)), rtac refl 1, 
    (rtac extrs_wkrls_adm 1), Fast_tac 1 ]) ;

(* ctr_adm_step for rule where principal formula copied into premises *)
val () = qed_goal_spec_mp "gen_ctr_adm_step_princ" lctr.thy 
  "(ps, c) : scrls --> (ALL p : set ps. c <= p) --> \
    \ epsc : extrs {(ps, c)} --> extrs {(ps, c)} <= drls --> \
    \ ctr_adm_step sub (derrec drls {}) epsc A"
  (fn _ => [ (clarify_tac (ces [extrs.elims]) 1),
    (rtac sc_ctr_adm_step2 1), TRYALL atac, rtac refl 1, (etac disjI2 1) ]) ;

val _ = qed_goal_spec_mp "mult_0_lem" lctr.thy "r = (rp # rps, rc) --> \
  \ (EX ps. c : set ps & (ps, c) : extrs (multiples {r}))" 
  (fn _ => [ Clarify_tac 1, rtac exI 1, rtac conjI 1,
    rtac extrs.I 2, (res_inst_tac [("n", "0")] multiples.I 2),
    (rtac singletonI 2), (force_tac (cesimps [extend_def]) 2),
    Simp_tac 1 ]) ;

*)

