
structure seq = struct val thy = theory "seq" end ;
use_legacy_bindings seq.thy ;

val extend_def' =
  [transfer seq.thy extend_def, mk_eq add_commute] MRS transitive_thm ;

val _ = bind_thm ("le_extend", 
  [extend_def RS def_imp_eq, le_plus1] MRS transfer seq.thy xt3) ;

val _ = qed_goal_spec_mp "extcs_leI" seq.thy 
  "(ps, c :: 'a sequent) : rls --> c <= c' --> (ps, c') : extcs rls"
  (fn _ => [ (safe_tac (ces [le_add_diff_exE])), 
    (rtac extcs.I 1), atac 1, (simp_tac (esimps [extend_def]) 1) ]) ;

(** for instance sequent :: (type) lattice **)

val _ = qed_goalw "seq_meet_pg" seq.thy 
  [pg_meet_def, ms_meet_pg, seq_meet_def, seq_minus_def]
  "seq_meet x y == pg_meet x y" fn_at ;

val _ = qed_goalw "seq_join_pg" seq.thy 
  [pg_join_def, ms_join_pg, seq_join_def, seq_minus_def, seq_plus_def]
  "seq_join x y == pg_join x y" fn_at ;

val seq_meet_alt = [pg_meet_alt, symmetric seq_meet_pg] MRS transitive_thm ;
val seq_join_alt = [pg_join_alt, symmetric seq_join_pg] MRS transitive_thm ;

val _ = qed_goalw "seq_inf" seq.thy [seq_meet_pg] 
  "is_inf x y (seq_meet x y)" fn_at;
val _ = qed_goalw "seq_sup" seq.thy [seq_join_pg] 
  "is_sup x y (seq_join x y)" fn_at;
AddIffs [seq_inf, seq_sup] ;

(* eg mk_seq meet_commute is "seq_meet ?x1 ?y1 = seq_meet ?y1 ?x1" *)
fun mk_seq th = rewrite_rule [seq_meet_alt, seq_join_alt]
  ([seq_eq_refl, transfer seq.thy th] MRS trans) ;
(*
val _ = bind_thm ("meet_seq", seq_inf RS meet_equality) ;
val _ = bind_thm ("join_seq", seq_sup RS join_equality) ;
*)

val _ = qed_goal "seq_delete_add" seq.thy 
  "M + seq_delete D XY >= seq_delete D (M + XY)"
  (fn _ => [ (simp_tac (esimps [seq_delete_sum]) 1) ]) ;

val _ = qed_goal "seq_times" seq.thy
  "times n (Sequent a s) = Sequent (times n a) (times n s)"
  (fn _ => [ (induct_tac "n" 1),
    (simp_tac (esimps [seq_zero_def]) 1), Asm_simp_tac 1] ) ;

(* when sequents are disjoint *)
val _ = qed_goalw "seq_disj_minus" seq.thy [seq_meet_pg]
  "(seq_meet c ic = 0) = (c - ic = c)" 
  (fn _ => [ (rtac pg_disj_minus 1) ]) ;

val _ = bind_thm ("seq_disj_le_add", seq_disj_minus RS iffD1 RS disj_le_add) ;

val ms_at_def' = rewrite_rule [atomic_def] ms_atomic_def RS sym ;

val () = qed_goalw "seq_atomic_def" seq.thy [atomic_def]
  "atomic (s :: 'a sequent) = (seq_size s <= 1)"
  (fn _ => [ (case_tac "s" 1), Asm_simp_tac 1,
    (case_tac "size multiset1" 1), 
    (case_tac "size multiset2" 2), 
    (ALLGOALS Asm_full_simp_tac),
    (res_inst_tac [("x", "multiset1 |- 0")] exI 3), 
    (asm_simp_tac (esimps [seq_zero_def, size_eq_0_iff_0 RS sym]) 3) ,
    (cut_inst_tac [("M1", "multiset2")] ms_at_def' 1) ,
    (cut_inst_tac [("M1", "multiset1")] ms_at_def' 2) ,
    (ALLGOALS (clarsimp_tac (cesimps [symmetric Zero_multiset_def,
      split_seq_all, seq_zero_def, le0_eq, ms_less_le]))) ]) ;

val _ = qed_goalw "seq_atomic_single" seq.thy [mk_eq seq_atomic_def]
  "atomic (0 |- {#A#}) & atomic ({#A#} |- 0)" fn_at ;
AddIffs (msc seq_atomic_single) ;
val seq_atomic_singles = msc seq_atomic_single ;

val _ = qed_goal_spec_mp "scrls_atomic" seq.thy "(ps, c) : scrls --> atomic c"
  (fn _ => [ (clarsimp_tac (ces [scrls.elims], esimps [seq_atomic_def]) 1) ]) ;

val _ = qed_goal_spec_mp "seq_atomic_ms" seq.thy
  "ms_of_seq As = {#x#} --> atomic As"
  (fn _ => [ (clarsimp_tac (cesimps [seq_atomic_def, seq_size_ms]) 1) ]);

val seq_atom_disj_or_le = 
  fold_rule [seq_meet_pg] (seq_atomic_def RS iffD2 RS atomicD) ;

val atom_le_union = 
  fold_rule [seq_meet_pg] (seq_atomic_def RS iffD2 RS atomic_le_sumD) ;
val atom_le_union' = rewrite_rule [One_nat_def] atom_le_union ;

val seq_atoms_disj_or_eq = fold_rule [seq_meet_pg]
  (tacthm (TRYALL (rtac (seq_atomic_def RS iffD2))) atoms_disj_or_eq) ;

(* weakening rules - conclusion atomic *)
val _ = qed_goal_spec_mp "wk_atomic" seq.thy 
  "((wkps, wkc) : wkrls) --> atomic wkc" 
  (fn _ => [ (simp_tac (esimps [wk_alt, seq_atomic_def]) 1) ]) ;

val _= qed_goal_spec_mp "ctr_double" seq.thy
  "(cps, cc) : ctrrls A --> cps = [cc + cc]"
  (fn _ => [ (safe_tac (ces [ctrrls.elims])), ALLGOALS Simp_tac ]) ;

(* contraction rules - difference atomic *)
val _= qed_goal_spec_mp "ctr_atomic_diff'" seq.thy
  "([cp], cc) : extrs (ctrrls A) --> cc <= cp & atomic (cp - cc)"
  (fn _ => [ (safe_tac (ces [extrs_E2, ctrrls.elims])),
    ALLGOALS Clarsimp_tac ]) ;

val _= qed_goal_spec_mp "ctr_atomic_diff" seq.thy
  "([cp], cc) : extrs (ctrrls A) --> (EX x. cp = cc + x & atomic x)"
  (fn _ => [ (rtac impI 1), (res_inst_tac [("x", "cp - cc")] exI 1),
    (safe_tac (ces [extrs_E2, ctrrls.elims])), (ALLGOALS Clarsimp_tac),
    (ALLGOALS (simp_tac (esimps add_ac addsimps [Zero_multiset_def]))) ]) ;

(* ric : thm -> thm
  apply conversion of union_lcomm *)
fun ric th = let val SOME resth = cvt (eqc union_lcomm) th in resth end 
  handle Bind => raise THM ("ric", 0, [th]) ;

val etacs = [ (rtac ext 1), (case_tac "x" 1), 
  (auto_tac (cesimps [extend_def, assoc]))] ;

val () = qed_goal "extend_comp" seq.thy 
  "extend S o extend S' = extend (S' + S)" (fn _ => etacs) ;

val extend_comp' = fcomm_all add_commute extend_comp ;

val () = qed_goal "extend_empty" seq.thy "extend 0 = id" (fn _ => etacs) ;

val [extend_empty', extend_comp_ext] = 
  map (simplify (simpset ())) ([extend_empty, extend_comp] RL [fun_cong]) ;

Addsimps [extend_empty] ; (* must follow defining extend_empty' *)

val _ = bind_thm ("extend_alt", 
  [transfer seq.thy extend_def, mk_eq add_commute] MRS transitive_thm) ;

(* 
set show_sorts ;
reset show_sorts ;
*)

val _ = qed_goal "extend_00" seq.thy "extend (0 |- 0) = id"
  (fn _ => etacs) ;

Addsimps [extend_00] ;

val _ = qed_goal_spec_mp "in_extrsl" seq.thy
  "(ps, c) : rules --> (ps, c) : extrsl rules"
  (fn _ => [ (clarify_tac (ces [extrsl.I]) 1),
  (instantiate_tac [("fl1","0"), ("fr1","0")]),
  (Simp_tac 1) ]) ;

val _ = qed_goal "extrsl_extrsl" seq.thy "extrsl (extrsl rls) = extrsl rls"
  (fn _ => [ (safe_tac (ces [in_extrsl, extrsl_E'])),
  (asm_full_simp_tac (HOL_ss addsimps
      [ extend_comp,
        map_compose RS sym,
        rewrite_rule [mk_eq o_apply] (extend_comp RS fun_cong) ]) 1),
  (simp_tac (esimps [seq_plus_def]) 1),
  (etac extrsl_I 1) ]) ;

val _ = bind_thm ("in_extrs", simplify (simpset ())
  (read_instantiate [("flr", "0")] extrs_I)) ;

val _ = qed_goal "extrs_extrs" seq.thy "extrs (extrs rls) = extrs rls"
  (fn _ => [ (safe_tac (ces [in_extrs])), (safe_tac (ces [extrs_E'])) , 
    (asm_full_simp_tac (HOL_ss addsimps
      [extend_comp, pscmap_compose RS sym]) 1),
    (etac extrs_I 1) ]) ;

val _ = bind_thm ("ext_extrs",
  [extrs_I, extrs_extrs RS equalityD1] MRS rev_subsetD) ;
val _ = bind_thm ("extrs_Ipc",
  [extrs_I, pscmap_def] MRS mem_same) ;

AddSEs [in_extrs, ext_extrs, extrs_I] ; 
Addsimps [extrs_extrs] ;

val _ = bind_thm ("in_iextrs", simplify (simpset ())
  (read_instantiate [("fl", "0")] iextrs_I)) ;

val _ = qed_goal "iextrs_iextrs" seq.thy "iextrs (iextrs rls) = iextrs rls"
  (fn _ => [ (safe_tac (ces [in_iextrs])), (safe_tac (ces [iextrs_E'])) , 
    (asm_full_simp_tac (HOL_ss addsimps
      [sds_plus_def, add_0, extend_comp, pscmap_compose RS sym]) 1),
    (etac iextrs_I 1) ]) ;

val _ = bind_thm ("ext_iextrs",
  [iextrs_I, iextrs_iextrs RS equalityD1] MRS rev_subsetD) ;
val _ = bind_thm ("iextrs_Ipc",
  [iextrs_I, pscmap_def] MRS mem_same) ;

AddSEs [in_iextrs, ext_iextrs, iextrs_I] ; 
Addsimps [iextrs_iextrs] ;

val _ = bind_thm ("in_extcs", simplify (simpset ())
  (read_instantiate [("flr", "0")] extcs_I)) ;

val _ = qed_goal "extcs_extcs" seq.thy "extcs (extcs rls) = extcs rls"
  (fn _ => [ (safe_tac (ces [in_extcs])), (safe_tac (ces [extcs_E'])) , 
    (simp_tac (HOL_ss addsimps 
      [rewrite_rule [o_def] extend_comp RS fun_cong]) 1),
    (etac extcs_I 1) ]) ;

val _ = bind_thm ("ext_extcs",
  [extcs_I, extcs_extcs RS equalityD1] MRS rev_subsetD) ;

AddSEs [in_extcs, ext_extcs, extcs_I] ; 
Addsimps [extcs_extcs] ;

(*
val iud = prove_goal seq.thy 
  "(X Un (insert A X - {A}) |- insert A Y - {A} Un Y) = (X |- Y)" fn_at ;

val () = qed_goal "Un_diffL" Set.thy "(A - B) Un A = A" fn_ft1 ;
val () = qed_goal "Un_diffR" Set.thy "A Un (A - B) = A" fn_ft1 ;
  *)

val prod_injectI = conjI RS (hd prod.inject RS iffD2) ;
val sequent_injectI = conjI RS (sequent.inject RS iffD2) ;

val icl = [union_lcomm, refl] MRS sequent_injectI ;
val icr = [refl, union_lcomm] MRS sequent_injectI ;
val iclms' = [refl, icl] MRS prod_injectI ;
val icrms' = [refl, icr] MRS prod_injectI ;
val icrmf' = [icr, refl] MRS prod_injectI ;
val [iclms, icrms, icrmf] = [iclms', icrms', icrmf'] RLN (2, [mem_same]) ;

(*
val iil = [insert_absorb2, refl] MRS sequent_injectI ;
val iir = [refl, insert_absorb2] MRS sequent_injectI ;
*)

val () = qed_goalw "extend_inj" seq.thy [extend_def]
  "(extend plr x = extend plr y) = (x = y)" fn_at ;

val () = qed_goalw "extend_inj'" seq.thy [extend_def]
  "(extend f x = extend p x) = (f = p)" fn_at ;

val () = qed_goal "lssl'" seq.thy
  "seq_size M = seq_size N --> M <= N --> M = N"
  (fn _ => [ Safe_tac, (etac order_leE 1),
    (dtac less_seq_size_less 1), Auto_tac]) ;
 
val lssl = reop rev (meta_std lssl') ;

val () = qed_goal_spec_mp "seq_size_Suc" seq.thy
  "seq_size M = Suc n --> (EX N X x. M = N + X & \
    \ seq_size N = n & ms_of_seq X = {#x#})"
  (fn _ => [ (case_tac "M" 1), 
    (case_tac "size multiset1" 1), 
    (case_tac "size multiset2" 1), 
    Asm_simp_tac 1, 
    (ALLGOALS (dtac size_eq_Suc_imp_elem)), Safe_tac,
    (ALLGOALS (rtac exI)),
    (res_inst_tac [("x", "(Sequent 0 {#a#})")] exI 1),
    (res_inst_tac [("x", "(Sequent {#a#} 0)")] exI 2),
    (ALLGOALS Simp_tac), Safe_tac,
    (TRYALL (rtac (le_add_diff_eqc RS sym))),
    (ALLGOALS (asm_full_simp_tac (esimps [single_le]))),
    (case_tac "size multiset1" 1), Auto_tac ]) ;

val _ = qed_goal "seq_singles_le" seq.thy "(seq_singles M <= M)"
  (fn _ => [ (case_tac "M" 1), Clarsimp_tac 1 ]) ;
Addsimps [seq_singles_le] ;

val _ = qed_goal_spec_mp "seq_singles_mono" seq.thy 
  "N <= M --> (seq_singles N <= seq_singles M)"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), Clarsimp_tac 1,
    (safe_tac (cis [set_of_mono, singles_mono, finite_set_of])) ]) ;

val _ = qed_goal "seq_singles_idemp" seq.thy 
  "(seq_singles (seq_singles M) = seq_singles M)"
  (fn _ => [ (case_tac "M" 1), Clarsimp_tac 1 ]) ;

val _ = qed_goal "seq_singles_double" seq.thy 
  "seq_singles (M + M) = seq_singles M"
  (fn _ => [ (case_tac "M" 1), Clarsimp_tac 1 ]) ;
Addsimps [seq_singles_idemp, seq_singles_double] ;
