
structure seqgen = struct val thy = theory "seqgen" end ;
use_legacy_bindings seqgen.thy ;

AddIffs [idrls.I] ; (* AddSIs [scrls.I] ; *)
val iscrls_s1I = scrls.I RS iscrls.s1I ;
val iscrls_idI = idrls.I RS iscrls.idI ;
AddIffs (wkrls.intros @ ctrrls.intros) ;
AddIffs (map (fold_rule [Zero_multiset_def]) (wkrls.intros @ ctrrls.intros)) ;
AddIffs [iscrls_idI] ; AddSEs [iscrls_s1I] ; 

val _ = qed_goal_spec_mp "scrlsD" seqgen.thy 
  "(ps, c) : scrls --> seq_size c = 1" 
  (fn _ => [ (safe_tac (ces [scrls.elims])) ]) ;

val _ = bind_thm ("scrls_iff", tacsthm [ etac scrlsD 1, etac scrls.I 1 ] iffI) ;

val ssaf = prove_goal seqgen.thy
  "(!!x. PROP P x) ==> (!!ant suc. PROP P (ant |- suc))" 
  (fn [p] => [ rtac p 1 ]) ;

val _ = qed_goal "antec_succ_simp" seqgen.thy
  "(antec seq |- succ seq) = seq"
  (fn _ => [(case_tac "seq" 1), Auto_tac ]) ;
Addsimps [antec_succ_simp] ;

val _ = qed_goal "split_seq_all" seqgen.thy
  "All P = (ALL ml mr. P (ml |- mr))"
  (fn _ => [ Safe_tac, (case_tac "x" 2), ALLGOALS Fast_tac ]) ;

val _ = qed_goal "split_seq_ex" seqgen.thy
  "Ex P = (EX ml mr. P (ml |- mr))"
  (fn _ => [ Safe_tac, etac exI 2, (case_tac "x" 1), Fast_tac 1 ]) ;

val ssab = prove_goal seqgen.thy
  "(!!ant suc. PROP P (ant |- suc)) ==> (!!x. PROP P x)" 
  (fn [p] => [ (cut_inst_tac [("ant", "antec x"), ("suc", "succ x")] p 1),
    (Full_simp_tac 1) ]) ;

val _ = bind_thm ("split_sequent_all", equal_intr ssaf ssab) ;
val split_sequent_tac = 
  safe_full_simp_tac (HOL_basic_ss addsimps [split_sequent_all]) ;

(* seqmap *)
val _ = qed_goal "antec_seqmap" seqgen.thy
  "antec (seqmap f seq) = mset_map f (antec seq)" (indAu "seq") ;

val _ = qed_goal "succ_seqmap" seqgen.thy
  "succ (seqmap f seq) = mset_map f (succ seq)" (indAu "seq") ;

val _ = qed_goal_spec_mp "seqmap_id" seqgen.thy "seqmap id = id" 
  (fn _ => [ (rtac ext 1), (case_tac "x" 1), 
    (clarsimp_tac (cesimps [mset_map_id]) 1) ]) ;

val _ = qed_goalw "seqmap_0" seqgen.thy [seq_zero_def]
  "seqmap f 0 = 0" fn_at ;

Addsimps [seqmap_0, seqmap_id, rewrite_rule [id_def] seqmap_id] ;

val _ = qed_goalw "seqmap_compose" seqgen.thy []
  "seqmap (f o g) x = seqmap f (seqmap g x)"
  (fn _ => [ (case_tac "x" 1),
    (clarsimp_tac (cesimps [mset_map_compose', o_def]) 1) ]) ;

val _ = qed_goalw "seqmap_compose'" seqgen.thy []
  "seqmap (f o g) = seqmap f o seqmap g"
  (fn _ => [ (rtac ext 1), (simp_tac (esimps [seqmap_compose]) 1) ]) ;
  
val wec = InductivePackage.the_mk_cases seqgen.thy "seqgen.wkrls" ;
val wkrls_np = wec "(ps, {#A#} |- {#A#}) : wkrls" ;
AddSEs [wkrls_np] ;

val _ = bind_thm ("iscrls_alt",
  simplify (HOL_ss addsimps [symmetric Un_def, lfp_const]) iscrls.defs) ;
    
(* prove axioms for sequent to be in type class plus_ac0, order, pm_ge0 *)

val _ = qed_goal "extend_fn" seqgen.thy "extend fmls = (%seq. seq + fmls)" 
  (fn _ => [ (rtac ext 1), (simp_tac (esimps [extend_def]) 1) ]) ;

val _ = bind_thm ("extrsI2", 
  read_instantiate [("flr", "(?fl |- ?fr)")] extrs.I) ;
val _ = bind_thm ("extcsI2", 
  read_instantiate [("flr", "(?fl |- ?fr)")] extcs.I) ;

val seq_defs = [seq_zero_def, seq_plus_def, seq_minus_def,
  seq_le_def, seq_less_def] ;
val sdss as [sds_extend_def, sds_plus_def, sds_minus_def,
  sds_le_def, sds_less_def] =
  map (simplify (simpset ()) o read_instantiate
    [("seq", "Sequent ?X ?Y"), ("fmls", "Sequent ?ant ?suc")]) 
    (extend_def :: tl seq_defs) ;
  
Addsimps sdss ;
Delsimps [seq_delete.simps] ;
val seq_delete_simp = simplify (esimps []) 
  (read_instantiate [("dseq", "(?dl |- ?dr)")] seq_delete.simps) ;
Addsimps [seq_delete_simp] ;

val sepdbl = (sds_plus_def RS def_imp_eq RS sym RSN (2, mem_same)) ;
val sepdbl' = (sds_plus_def RS def_imp_eq RSN (2, mem_same)) ;

val seq_eq_refl = prove_goal seqgen.thy "(a :: 'a sequent) = a" fn_at ;

val _ = qed_goal_spec_mp "seq_eqI" seqgen.thy 
  "antec seqa = antec seqb --> succ seqa = succ seqb --> seqa = seqb"
  (fn _ => [ (case_tac "seqa" 1), (case_tac "seqb" 1), Force_tac 1 ]) ;

val _ = qed_goal_spec_mp "seq_eq_iff" seqgen.thy 
  "(seqa = seqb) = (antec seqa = antec seqb & succ seqa = succ seqb)"
  (fn _ => [ (fast_tac (cis [seq_eqI]) 1) ]) ;

val _ = qed_goalw_spec_mp "antec_sum" seqgen.thy [seq_plus_def]
  "antec (seqa + seqb) = antec seqa + antec seqb" fn_at ;
val _ = qed_goalw_spec_mp "succ_sum" seqgen.thy [seq_plus_def]
  "succ (seqa + seqb) = succ seqa + succ seqb" fn_at ;

val _ = qed_goalw_spec_mp "antec_diff" seqgen.thy [seq_minus_def]
  "antec (seqa - seqb) = antec seqa - antec seqb" fn_at ;
val _ = qed_goalw_spec_mp "succ_diff" seqgen.thy [seq_minus_def]
  "succ (seqa - seqb) = succ seqa - succ seqb" fn_at ;

val _ = qed_goal_spec_mp "seqmap_mono" seqgen.thy 
  "A <= B --> seqmap f A <= seqmap f B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), 
    (clarsimp_tac (cesimps [mset_map_mono]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_plus" seqgen.thy 
  "seqmap f (A + B) = seqmap f A + seqmap f B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), 
    (clarsimp_tac (cesimps [mset_map_union]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_inj" seqgen.thy 
  "inj f --> (seqmap f M = seqmap f N) = (M = N)"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), 
    (clarsimp_tac (cesimps [mset_map_inj]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_diff_le" seqgen.thy 
  "B <= A --> seqmap f (A - B) = seqmap f A - seqmap f B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), 
    (clarsimp_tac (cesimps [mset_map_diff_le]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_diff_inj" seqgen.thy 
  "inj f --> seqmap f (A - B) = seqmap f A - seqmap f B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), 
    (clarsimp_tac (cesimps [mset_map_diff_inj]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_le_inj" seqgen.thy
  "inj f --> seqmap f A <= seqmap f B --> A <= B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), 
    (clarsimp_tac (cesimps [mset_map_le_inj]) 1) ]) ;

val _ = qed_goal_spec_mp "le_seqmap" seqgen.thy
  "Y <= seqmap f X --> (EX Y'. Y' <= X & Y = seqmap f Y')"
  (fn _ => [ (case_tac "X" 1), (case_tac "Y" 1), 
    (force_tac (cds [le_mset_map], esimps [split_seq_ex]) 1) ]) ;

val _ = qed_goal_spec_mp "seqmap_split" seqgen.thy
  "seqmap f X' = Y + Z --> (EX Y' Z'. X' = Y' + Z' & \
    \ seqmap f Y' = Y & seqmap f Z' = Z)"
  (fn _ => [ (case_tac "X'" 1), (case_tac "Y" 1), (case_tac "Z" 1), 
    (force_tac (cds [mset_map_split], esimps [split_seq_ex, le_mset_map]) 1) ]) ;

val pm_ths = [diff_0_le, pm0_class.pm0_minus_minus, le_add_diff_eq] ;
val ttcs = [ (case_tac_frees 1), TRY Safe_tac,
  (ALLGOALS (asm_full_simp_tac (esimps (
    symmetric Zero_multiset_def :: seq_zero_def :: pm_ths @ union_ac))))] ;

val extrs_I = refl RSN (2, extrs.I) ;
val extrs_E = tacthm (hyp_subst_tac 2) extrs.elims ;
val extrs_E' = tacthm (dtac sym 2) extrs_E ;

val iextrs_I = refl RSN (2, iextrs.I) ;
val iextrs_E = tacthm (hyp_subst_tac 2) iextrs.elims ;
val iextrs_E' = tacthm (dtac sym 2) iextrs_E ;

val extcs_I = refl RSN (2, extcs.I) ;
val extcs_E = tacthm (hyp_subst_tac 2) extcs.elims ;
val extcs_E' = tacthm (dtac sym 2) extcs_E ;

(* TODO *)

val extrsl_I = refl RSN (2, extrsl.I) ;
val extrsl_E = tacthm (hyp_subst_tac 2) extrsl.elims ;
val extrsl_E' = tacthm (dtac sym 2) extrsl_E ;

(* TODO *)

val _ = qed_goal_spec_mp "extrs_D2" seqgen.thy  
  "epsc : extrs rules --> \
    \ (EX psc fl fr. psc : rules & pscmap (extend (fl |- fr)) psc = epsc)"
  (fn _ => [ (clarify_tac (ces [extrs_E]) 1), (case_tac "flr" 1), Fast_tac 1 ]);

val extrs_E2 = tacthm (REPEAT (eresolve_tac [exE, conjE] 2))
  (extrs_D2 RS revcut_rl_bool) ;
val extrs_E2' = tacthm (dtac sym 2) extrs_E2 ;

val _ = qed_goal_spec_mp "extcs_D2" seqgen.thy  
  "epsc : extcs rules --> \
    \ (EX ps c fl fr. (ps, c) : rules & (ps, extend (fl |- fr) c) = epsc)"
  (fn _ => [ (clarify_tac (ces [extcs_E]) 1), (case_tac "flr" 1), Fast_tac 1 ]);

val extcs_E2 = tacthm (REPEAT (eresolve_tac [exE, conjE] 2))
  (extcs_D2 RS revcut_rl_bool) ;
val extcs_E2' = tacthm (dtac sym 2) extcs_E2 ;

val _ = qed_goal_spec_mp "extrs_mono" seqgen.thy
  "A <= B --> (extrs A) <= extrs B"
  (fn _ => [ (safe_tac (ces [extrs_E])), (dtac sym 1),
    (asm_simp_tac HOL_ss 1), (rtac extrs_I 1), Fast_tac 1 ]) ;

val extrs_mono' = mk_mono' extrs_mono ;

val _ = bind_thm ("extrs_single_sub",
  singleton_subsetI RS subset_trans RS extrs_mono) ;

val _ = bind_thm ("extrs_single", singleton_subsetI RS extrs_mono) ;
val _ = bind_thm ("extrs_1_sub", extrs_single RS subsetD) ;

val _ = qed_goal "extrs_UN" seqgen.thy 
  "(UN a:A. extrs (f a)) = extrs (UN a:A. f a)"
  (fn _ => [ (safe_tac (ces [extrs_E])), (etac UN_I 2), 
    (ALLGOALS (rtac extrs.I)), TRYALL atac, Fast_tac 1 ]) ;

val _ = qed_goal "extrs_UN_single" seqgen.thy 
  "(UN a:A. extrs {a}) = extrs A"
  (fn _ => [ (simp_tac (esimps [extrs_UN]) 1) ]) ;

val _ = qed_goal_spec_mp "extrs_singleD" seqgen.thy 
  "x : extrs rls --> (EX y: rls. x : extrs {y})"
  (fn _ => [ (safe_tac (ces [extrs_E])), (etac rev_bexI 1),
    (rtac (singletonI RS extrs_I) 1) ]) ;

val extrs_singleE = tacthm (split_all_tac 2) (extrs_singleD RS bexE) ;

val _ = qed_goal "extrs_Un" seqgen.thy 
  "extrs (A Un B) = (extrs A Un extrs B)"
  (fn _ => [ Safe_tac, (TRYALL (fast_tac (ces [mk_mono' extrs_mono]))),
    (etac extrs_singleE 1), (etac UnE 1), etac notE 2, 
    (ALLGOALS (eatac (mk_mono' (singleton_subsetI RS extrs_mono)) 1)) ]) ;

val _ = qed_goal_spec_mp "extcs_mono" seqgen.thy
  "A <= B --> (extcs A) <= extcs B"
  (fn _ => [ (safe_tac (ces [extcs_E])),
    (datac subsetD 1 1), (etac extcs_I 1) ]) ;

val extcs_mono' = mk_mono' extcs_mono ;

val _ = bind_thm ("extcs_single_sub",
  singleton_subsetI RS subset_trans RS extcs_mono) ;

val _ = bind_thm ("extcs_single", singleton_subsetI RS extcs_mono) ;
val _ = bind_thm ("extcs_1_sub", extcs_single RS subsetD) ;
val _ = qed_goal "extcs_UN" seqgen.thy 
  "(UN a:A. extcs (f a)) = extcs (UN a:A. f a)"
  (fn _ => [ (safe_tac (ces [extcs_E])), (etac UN_I 2), 
    (ALLGOALS (rtac extcs.I)), (TRYALL (ares_tac [refl])), Fast_tac 1 ]) ;

val _ = qed_goal "extcs_UN_single" seqgen.thy 
  "(UN a:A. extcs {a}) = extcs A"
  (fn _ => [ (simp_tac (esimps [extcs_UN]) 1) ]) ;

val _ = qed_goal_spec_mp "extcs_singleD" seqgen.thy 
  "x : extcs rls --> (EX y: rls. x : extcs {y})"
  (fn _ => [ (safe_tac (ces [extcs_E])), (etac rev_bexI 1),
    (rtac (singletonI RS extcs_I) 1) ]) ;

val extcs_singleE = tacthm (split_all_tac 2) (extcs_singleD RS bexE) ;

val _ = qed_goal "extcs_Un" seqgen.thy 
  "extcs (A Un B) = (extcs A Un extcs B)"
  (fn _ => [ Safe_tac, (TRYALL (fast_tac (ces [mk_mono' extcs_mono]))),
    (etac extcs_singleE 1), (etac UnE 1), etac notE 2, 
    (ALLGOALS (eatac (mk_mono' (singleton_subsetI RS extcs_mono)) 1)) ]) ;

val _ = qed_goal "seq_commute" seqgen.thy "x + y = y + (x :: 'a sequent)" 
  (fn _ => ttcs) ;
val _ = qed_goal "seq_assoc" seqgen.thy 
  "(x + y) + z = (x :: 'a sequent) + (y + z)" (fn _ => ttcs) ;
val _ = qed_goal "seq_zero" seqgen.thy "0 + x = (x :: 'a sequent)" 
  (fn _ => ttcs) ;

val _ = qed_goal "seq_refl" seqgen.thy "x <= (x :: 'a sequent)" (fn _ => ttcs) ;
val _ = qed_goal_spec_mp "seq_trans" seqgen.thy
  "x <= y --> y <= z --> (x :: 'a sequent) <= z" 
  (fn _ => ttcs @ [Safe_tac, REPEAT (eatac order_trans 1 1)]) ;
val _ = qed_goal_spec_mp "seq_antisym" seqgen.thy
  "x <= y --> y <= x --> (x :: 'a sequent) = y"
  (fn _ => ttcs @ [Safe_tac, REPEAT (eatac order_antisym 1 1)]) ;
val _ =  bind_thm ("seq_less_le", seq_less_def RS def_imp_eq) ;

val _ = bind_thm ("Pair_antec_succ_simp", 
  [Pair_eq, [antec_succ_simp, antec_succ_simp] MRS conjI] MRS iffD2) ;

val _ = bind_thm ("concl_antec_succ_simp", 
  [Pair_eq, [refl, antec_succ_simp] MRS conjI] MRS iffD2) ;

fun simpWZ th =
  let val th1 = read_instantiate [("W", "antec ?WZ"), ("Z", "succ ?WZ")] th ;
  in rewrite_rule [mk_eq antec_succ_simp] th1 end ;

fun simpUV th =
  let val th1 = read_instantiate [("U", "antec ?UV"), ("V", "succ ?UV")] th ;
  in rewrite_rule [mk_eq antec_succ_simp] th1 end ;

val _ = qed_goal "seq_all_ge0" seqgen.thy
  "0 <= (A :: 'a sequent)" (fn _ => ttcs) ;
val _ = qed_goal "seq_diff_0_le" seqgen.thy
  "(A - B = 0) = ((A :: 'a sequent) <= B)" (fn _ => ttcs) ;
val _ = qed_goal "seq_plus_minus" seqgen.thy
  "A + B - (A :: 'a sequent) = B" (fn _ => ttcs) ;
val _ = qed_goal "seq_minus_minus" seqgen.thy
  "A - B - C = (A :: 'a sequent) - (B + C)" (fn _ => ttcs) ;
val _ = qed_goal_spec_mp "seq_le_add_diff_eq" seqgen.thy 
  "B <= A --> B + ((A :: 'a sequent) - B) = A" (fn _ => ttcs) ;
val _ = qed_goal_spec_mp "seq_add_left_mono" seqgen.thy 
  "A <= B --> C + (A :: 'a sequent) <= C + B" (fn _ => ttcs) ;
val _ = qed_goal_spec_mp "seq_add_le_imp_le_left" seqgen.thy 
  "C + A <= C + B --> (A :: 'a sequent) <= B" (fn _ => ttcs) ;
val _ = qed_goal_spec_mp "seq_add_imp_eq" seqgen.thy 
  "C + A = C + B --> (A :: 'a sequent) = B" (fn _ => ttcs) ;

val _ = qed_goal_spec_mp "sum_le_1" HOL_Gen.thy
  "a + b <= Suc 0 --> a = 0 | b = 0" (fn _ => [ arith_tac 1 ]) ;

(* consequence of atomicity *)

val () = qed_goal_spec_mp "atom_a_or_s" seqgen.thy
  "seq_size seq = 1 --> (EX x. seq = ({#x#} |- {#}) | seq = ({#} |- {#x#}))"
  (fn _ => [ (case_tac "seq" 1),
    (case_tac "size multiset1" 1), Auto_tac, 
    (ALLGOALS (EVERY' [etac rev_mp, rtac multiset_induct])), Auto_tac ]) ;

val atom_a_or_s' = rewrite_rule [One_nat_def] atom_a_or_s ;

(* characterisation of weakening rules *)
val _ = qed_goal "wk_alt" seqgen.thy
  "((wkps, wkc) : wkrls) = (wkps = [0] & seq_size wkc = 1)"
  (fn _ => [ Safe_tac, (dtac atom_a_or_s 3),
    (TRYALL (etac wkrls.elims)), Auto_tac ]) ;

val _ = bind_thm ("wk_altE", wk_alt RS iffD1 RS conjE) ;

val () = qed_goal_spec_mp "less_seq_size_less" seqgen.thy 
  "M < N --> seq_size M < seq_size N"
  (fn _ => [ (case_tac_frees 1), (asm_simp_tac (esimps seq_defs) 1), 
    Safe_tac, (ALLGOALS (datac order_lessI 1)),
    (ALLGOALS (dtac less_size_less THEN' dtac le_size_le)),
    (ALLGOALS arith_tac)]) ;

val () = qed_goal "seq_size_ms" seqgen.thy 
  "seq_size A = size (ms_of_seq A)" (fn _ => ttcs) ;

val () = qed_goalw "seq_size_eq0" seqgen.thy [seq_zero_def]
  "(seq_size A = 0) = (A = 0)" (fn _ => ttcs) ;

val () = qed_goal "seq_size_plus" seqgen.thy
  "seq_size (M + N) = seq_size M + seq_size N" (fn _ => ttcs) ;

val () = qed_goalw "ms_of_seq_eq0" seqgen.thy [seq_zero_def] 
  "(ms_of_seq A = 0) = (A = 0)" (fn _ => ttcs) ;

val () = qed_goal "ms_of_seq_plus" seqgen.thy
  "ms_of_seq (M + N) = ms_of_seq M + ms_of_seq N" (fn _ => ttcs) ;

val () = ListPair.app bind_thm (["seq_size_0", "ms_of_seq_0"],
  [refl] RL ([seq_size_eq0, ms_of_seq_eq0] RL [iffD2]));
 
val () = qed_goal "seq_size_map" seqgen.thy
  "seq_size (seqmap f seq) = seq_size seq"
  (fn _ => [ (case_tac "seq" 1), Clarsimp_tac 1 ]) ;

Addsimps [seq_size_0, seq_size_eq0, seq_size_plus, seq_size_map,
  ms_of_seq_0, ms_of_seq_eq0, ms_of_seq_plus] ;

val _ = qed_goal "ms_of_seqmap" seqgen.thy
  "ms_of_seq (seqmap f seq) = mset_map f (ms_of_seq seq)"
  (fn _ => [ (case_tac "seq" 1), (clarsimp_tac (cesimps [mset_map_union]) 1)]); 

val _ = qed_goal_spec_mp "ms_of_seq_mono" seqgen.thy
  "A <= B --> ms_of_seq A <= ms_of_seq B"
  (fn _ => [ (case_tac "A" 1), (case_tac "B" 1), Clarsimp_tac 1,
    (eatac add_mono 1 1) ]) ;

val _ = qed_goal_spec_mp "ms_of_single" seqgen.thy
  "ms_of_seq seq = {#A#} --> seq = ({#A#} |- {#}) | seq = ({#} |- {#A#})"
  (fn _ => [ (rtac impI 1), (subgoal_tac "seq_size seq = 1" 1),
    (asm_simp_tac (esimps [seq_size_ms]) 2),
    (dtac atom_a_or_s 1), Auto_tac ]) ;

val _ = qed_goalw_spec_mp "seq_filter_mono" seqgen.thy [seq_le_def]
  "(ALL a. p a --> q a) --> seq_filter p M <= seq_filter q M"
  (fn _ => [ (case_tac "M" 1), Clarsimp_tac 1,
    Safe_tac, (TRYALL (etac ms_filter_mono)), Auto_tac ]) ;

val _ = qed_goal_spec_mp "seq_delete_mono" seqgen.thy 
  "dseq <= eseq --> N <= M --> seq_delete eseq N <= seq_delete dseq M"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1),
    (case_tac "eseq" 1), (case_tac "dseq" 1), 
    (force_tac (cis [ms_delete_mono, set_of_mono], esimps []) 1) ]) ;

val _ = qed_goal "seq_filter_sum" seqgen.thy 
  "seq_filter p (M + N) = seq_filter p M + seq_filter p N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), 
    (clarsimp_tac (cesimps [ms_filter_sum]) 1) ]);

val _ = qed_goal "seq_delete_sum" seqgen.thy 
  "seq_delete dseq (M + N) = seq_delete dseq M + seq_delete dseq N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), (case_tac "dseq" 1), 
    (asm_simp_tac (esimps [ms_delete_sum]) 1) ]) ;

val _ = qed_goal "seq_filter_diff1" seqgen.thy 
  "seq_filter p (M - N) = seq_filter p M - seq_filter p N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), 
    (clarsimp_tac (cesimps [ms_filter_diff1]) 1) ]);

val _ = qed_goal "seq_delete_diff1" seqgen.thy 
  "seq_delete dseq (M - N) = seq_delete dseq M - seq_delete dseq N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), (case_tac "dseq" 1), 
    (asm_simp_tac (esimps [ms_delete_diff1]) 1) ]) ;

val _ = qed_goal "seq_filter_diff2" seqgen.thy 
  "seq_filter p (M - N) = seq_filter p M - N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), 
    (clarsimp_tac (cesimps [ms_filter_diff2]) 1) ]);

val _ = qed_goal "seq_delete_diff2" seqgen.thy 
  "seq_delete dseq (M - N) = seq_delete dseq M - N"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), (case_tac "dseq" 1), 
    (asm_simp_tac (esimps [ms_delete_diff2]) 1) ]) ;

val _ = qed_goalw "seq_delete_empty" seqgen.thy [seq_zero_def]
  "seq_delete 0 M = M" (fn _ => [ (case_tac "M" 1), Asm_simp_tac 1 ]) ;

val _ = qed_goalw "seq_delete_self" seqgen.thy [seq_zero_def]
  "seq_delete M M = 0" (fn _ => [ (case_tac "M" 1), Asm_simp_tac 1 ]) ;

val _ = qed_goal "seq_delete_le" seqgen.thy "seq_delete N M <= M"
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1), Asm_simp_tac 1 ]) ;

Addsimps [seq_delete_empty, seq_delete_self, seq_delete_le] ;

val _ = qed_goal_spec_mp "seq_delete_delete" seqgen.thy 
  "seq_delete S (seq_delete T M) = seq_delete (S + T) M"
  (fn _ => [ (case_tac "M" 1), (case_tac "S" 1), (case_tac "T" 1),
    Asm_simp_tac 1 ]) ;

val _ = qed_goal_spec_mp "seq_delete_disj" seqgen.thy 
  "(seq_delete N M = M) = (M - N = M)" 
  (fn _ => [ (case_tac "M" 1), (case_tac "N" 1),
    (asm_simp_tac (esimps [ms_delete_disj]) 1) ]) ;

